"""Testcases for VisT dataset mapper."""
import unittest

import torch
from detectron2.config import get_cfg
from detectron2.data import DatasetFromList
from scalabel.label.typing import Frame

from vist.struct import Images, InputSample

from .dataset_mapper import (
    DataloaderConfig,
    DatasetMapper,
    MapDataset,
    ReferenceSamplingConfig,
)


class TestDatasetMapper(unittest.TestCase):
    """DatasetMapper Testcase class."""

    def test_reference_sampling(self) -> None:
        """Testcase for reference view sampling."""
        cfg = ReferenceSamplingConfig(
            type="sequential", num_ref_imgs=2, scope=3
        )

        data_dict = [
            dict(video_name=str(i % 2), frame_index=i - i // 2 - i % 2)
            for i in range(200)
        ]
        mapper = MapDataset(cfg, True, DatasetFromList(data_dict), lambda x: x)

        idcs = mapper.sample_ref_idcs(str(0), 50)
        self.assertTrue(idcs == [52, 54])

        idcs = mapper.sample_ref_idcs(str(0), 196)
        self.assertTrue(idcs == [194, 198])

    def test_getitem_fallback(self) -> None:
        """Testcase for getitem fallback if None is returned."""
        cfg = ReferenceSamplingConfig(
            type="sequential", num_ref_imgs=2, scope=3
        )

        data_dict = [
            dict(video_name=i % 2, frame_index=i - i // 2 - i % 2)
            for i in range(200)
        ]
        mapper = MapDataset(
            cfg, True, DatasetFromList(data_dict), lambda x: None
        )
        self.assertRaises(ValueError, mapper.__getitem__, 0)

    def test_transform_annotations(self) -> None:
        """Test the transform annotations method in DatasetMapper."""
        cfg = get_cfg()
        loader_cfg = DataloaderConfig(
            workers_per_gpu=0,
            image_channel_mode="BGR",
            ref_sampling_cfg=ReferenceSamplingConfig(num_ref_imgs=1, scope=1),
        )
        ds_mapper = DatasetMapper(loader_cfg, cfg)
        input_sample = InputSample(
            Frame(name="0"),
            Images(torch.zeros(1, 3, 128, 128), [(128, 128)]),
        )
        ds_mapper.transform_annotation(input_sample, None, lambda x: x)
        self.assertEqual(len(input_sample.boxes2d), 0)
        ds_mapper.transform_annotation(input_sample, [], lambda x: x)
        self.assertEqual(len(input_sample.boxes2d), 0)

    def test_sort_samples(self) -> None:
        """Test the sort_samples method in MapDataset."""
        cfg = ReferenceSamplingConfig(
            num_ref_imgs=1, scope=1, frame_order="temporal"
        )
        data_dict = [
            dict(video_name=i % 2, frame_index=i - i // 2 - i % 2)
            for i in range(200)
        ]
        ds_mapper = MapDataset(
            cfg, True, DatasetFromList(data_dict), lambda x: None
        )
        input_samples = [
            InputSample(
                Frame(name="1", frame_index=1),
                Images(torch.zeros(1, 3, 128, 128), [(128, 128)]),
            ),
            InputSample(
                Frame(name="0", frame_index=0),
                Images(torch.zeros(1, 3, 128, 128), [(128, 128)]),
            ),
        ]

        sorted_samples = ds_mapper.sort_samples(input_samples)
        self.assertEqual(sorted_samples[0].metadata.frame_index, 0)
        self.assertEqual(sorted_samples[1].metadata.frame_index, 1)



"""Load MOTChallenge format dataset into Scalabel format."""
import inspect
import os

from scalabel.label.from_mot import from_mot
from scalabel.label.io import load_label_config
from scalabel.label.typing import Dataset

from .base import BaseDatasetLoader


class MOTChallenge(BaseDatasetLoader):
    """Custom dataloading class."""

    def load_dataset(self) -> Dataset:  # pragma: no cover
        """Convert MOTChallenge annotations to scalabel format."""
        assert (
            self.cfg.data_root == self.cfg.annotations
        ), "MOTChallenge requires images and annotations in the same path."
        cfg_path = self.cfg.config_path
        if cfg_path is None:
            cfg_path = os.path.join(
                os.path.dirname(os.path.abspath(inspect.stack()[1][1])),
                "motchallenge.toml",
            )
        metadata_cfg = load_label_config(cfg_path)

        frames = from_mot(self.cfg.annotations)
        return Dataset(frames=frames, config=metadata_cfg)



"""Base class for VisT models."""

import abc
from typing import Dict, List, Tuple, Union

import torch
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field

from vist.common.registry import RegistryHolder
from vist.struct import Boxes2D, InputSample, LossesType, ModelOutput


class BaseModelConfig(PydanticBaseModel, extra="allow"):
    """Config for default VisT tracker."""

    type: str = Field(...)


class BaseModel(torch.nn.Module, metaclass=RegistryHolder):  # type: ignore
    """Base tracker class."""

    @property
    @abc.abstractmethod
    def device(self) -> torch.device:
        """Get device where input should be moved to."""
        raise NotImplementedError

    def forward(
        self, batch_inputs: List[List[InputSample]]
    ) -> Union[LossesType, ModelOutput]:
        """Model forward function."""
        if self.training:
            return self.forward_train(batch_inputs)
        inputs = [inp[0] for inp in batch_inputs]  # no ref views during test
        return self.forward_test(inputs)

    @abc.abstractmethod
    def forward_train(
        self, batch_inputs: List[List[InputSample]]
    ) -> LossesType:
        """Forward pass during training stage.

        Args:
            batch_inputs: Model input. Batched, including possible reference
            views.

        Returns:
            LossesType: A dict of scalar loss tensors.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def forward_test(
        self, batch_inputs: List[InputSample], postprocess: bool = True
    ) -> ModelOutput:
        """Forward pass during testing stage.

        Args:
            batch_inputs: Model input (batched).
            postprocess: If output should be postprocessed to original
            resolution.

        Returns:
            ModelOutput: Dict of LabelInstance results, e.g. tracking and
            separate models result.
        """
        raise NotImplementedError

    @staticmethod
    def postprocess(
        original_wh: Tuple[int, int],
        output_wh: Tuple[int, int],
        detections: Boxes2D,
    ) -> None:
        """Postprocess results."""
        scale_factor = (
            original_wh[0] / output_wh[0],
            original_wh[1] / output_wh[1],
        )
        detections.scale(scale_factor)
        detections.clip(original_wh)


def build_model(cfg: BaseModelConfig) -> BaseModel:
    """Build VisT model.

    Note that it does not load any weights from ``cfg``.
    """
    assert cfg is not None
    registry = RegistryHolder.get_registry(BaseModel)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseModel)
        return module
    raise NotImplementedError(f"Model {cfg.type} not found.")



"""Base class for meta architectures."""

import abc
from typing import Any, Optional

import torch
from pydantic import BaseModel, Field

from vist.common.registry import RegistryHolder


class LossConfig(BaseModel, extra="allow"):
    """Base loss config."""

    type: str = Field(...)
    reduction: str = "mean"
    loss_weight: Optional[float] = 1.0


class BaseLoss(torch.nn.Module, metaclass=RegistryHolder):  # type: ignore
    """Base loss class."""

    @abc.abstractmethod
    def forward(  # type: ignore
        self, *args: Any, **kwargs: Any
    ) -> torch.Tensor:
        """Loss function implementation.

        Returns the reduced loss (scalar).
        """
        raise NotImplementedError


def build_loss(cfg: LossConfig) -> BaseLoss:
    """Build the loss functions for detect."""
    registry = RegistryHolder.get_registry(BaseLoss)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseLoss)
        return module
    raise NotImplementedError(f"Loss function {cfg.type} not found.")



"""Load and convert bdd100k labelsd to scalabel format."""

from bdd100k.common.utils import load_bdd100k_config
from bdd100k.label.to_scalabel import bdd100k_to_scalabel
from scalabel.label.io import load
from scalabel.label.typing import Dataset

from .base import BaseDatasetLoader


class BDD100K(BaseDatasetLoader):
    """BDD100K dataloading class."""

    def load_dataset(self) -> Dataset:
        """Convert BDD100K annotations to Scalabel format and prepare them."""
        assert self.cfg.annotations is not None
        bdd100k_anns = load(
            self.cfg.annotations,
            validate_frames=self.cfg.validate_frames,
            nprocs=self.cfg.num_processes,
        )
        frames = bdd100k_anns.frames
        assert self.cfg.config_path is not None
        bdd100k_cfg = load_bdd100k_config(self.cfg.config_path)

        scalabel_frames = bdd100k_to_scalabel(frames, bdd100k_cfg)
        return Dataset(frames=scalabel_frames, config=bdd100k_cfg.scalabel)



"""Backends for the data types a dataset of interest is saved in."""
from abc import abstractmethod

from pydantic import BaseModel

from vist.common.registry import RegistryHolder


class DataBackendConfig(BaseModel, extra="allow"):
    """Base data backend config."""

    type: str = "FileBackend"


class BaseDataBackend(metaclass=RegistryHolder):
    """Abstract class of storage backends.

    All backends need to implement two apis: ``get()`` and ``get_text()``.
    ``get()`` reads the file as a byte stream and ``get_text()`` reads the file
    as texts.
    """

    @abstractmethod
    def get(self, filepath: str) -> bytes:
        """Get the file content as bytes."""
        raise NotImplementedError


def build_data_backend(cfg: DataBackendConfig) -> BaseDataBackend:
    """Build a data backend from config."""
    registry = RegistryHolder.get_registry(BaseDataBackend)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseDataBackend)
        return module
    raise NotImplementedError(f"Data backend {cfg.type} not found.")



"""Visualizer class."""
import copy
import os
from collections import defaultdict
from typing import Dict, List

import torch
from detectron2.data import MetadataCatalog
from detectron2.evaluation import DatasetEvaluator
from detectron2.utils import comm
from PIL import Image
from scalabel.label.io import group_and_sort, save
from scalabel.label.typing import Frame

from ..common.utils.parallel import gather_predictions
from ..struct import Boxes2D, InputSample, ModelOutput
from .track import draw_sequence


class ScalabelVisualizer(DatasetEvaluator):  # type: ignore
    """Run model on sequence and visualize & save output."""

    def __init__(
        self,
        dataset_name: str,
        output_dir: str,
        distributed: bool = True,
        visualize: bool = True,
    ) -> None:
        """Init."""
        self._distributed = distributed
        self._output_dir = output_dir
        self._metadata = MetadataCatalog.get(dataset_name)
        self._predictions = defaultdict(list)  # type: Dict[str, List[Frame]]
        self._visualize = visualize

    def reset(self) -> None:
        """Preparation for a new round of evaluation."""
        self._predictions = defaultdict(list)

    def process(
        self, inputs: List[List[InputSample]], outputs: ModelOutput
    ) -> None:
        """Process the pair of inputs and outputs."""
        for key, output in outputs.items():
            for inp, out in zip(inputs, output):
                prediction = copy.deepcopy(inp[0].metadata)
                boxes2d = out.to(torch.device("cpu"))
                assert isinstance(
                    boxes2d, Boxes2D
                ), "Only Boxes2D output support for visualization."
                prediction.labels = boxes2d.to_scalabel(
                    self._metadata.idx_to_class_mapping
                )
                boxes2d.metadata = {
                    str(k): v
                    for k, v in self._metadata.idx_to_class_mapping.items()
                }
                attr = (
                    prediction.attributes
                    if prediction.attributes is not None
                    else dict()
                )
                attr["boxes2d"] = boxes2d  # type: ignore
                prediction.attributes = attr
                self._predictions[key].append(prediction)

    def evaluate(self) -> None:
        """Evaluate the performance after processing all input/output pairs."""
        if self._distributed:
            predictions_dict = gather_predictions(self._predictions)
            if not comm.is_main_process():
                return  # pragma: no cover
        else:
            predictions_dict = self._predictions  # pragma: no cover

        os.makedirs(os.path.join(self._output_dir), exist_ok=True)
        for key, predictions in predictions_dict.items():
            # save predictions
            predictions_without_boxes2d = []
            for frame in predictions:
                frame_without_boxes2d = copy.deepcopy(frame)
                assert frame_without_boxes2d.attributes is not None
                frame_without_boxes2d.attributes.pop("boxes2d")
                predictions_without_boxes2d.append(frame_without_boxes2d)

            save(
                os.path.join(self._output_dir, f"{key}_predictions.json"),
                predictions_without_boxes2d,
            )

            # visualize predictions
            if self._visualize:
                has_videos = True
                for frame in predictions:
                    if frame.video_name is None:
                        has_videos = False

                if has_videos:
                    predictions_grouped = group_and_sort(predictions)
                else:
                    predictions_grouped = [predictions]
                for video_predictions in predictions_grouped:
                    images = [
                        Image.open(frame.url) for frame in video_predictions
                    ]
                    boxes2d = []  # type: List[Boxes2D]
                    image_paths = []
                    for frame in video_predictions:
                        assert frame.attributes is not None
                        boxes2d.append(frame.attributes["boxes2d"])  # type: ignore # pylint: disable=line-too-long
                        if frame.video_name is None:
                            frame.video_name = ""
                        image_paths.append(
                            os.path.join(
                                self._output_dir,
                                f"{key}_visualization",
                                frame.video_name,
                                frame.name,
                            )
                        )
                    for fp, img in zip(
                        image_paths, draw_sequence(images, boxes2d)
                    ):
                        os.makedirs(os.path.dirname(fp), exist_ok=True)
                        img.save(fp)



"""Testcases for combined sampler."""
import unittest
from typing import List, Tuple

import torch

from vist.struct import Boxes2D

from ..matchers.base import MatchResult
from .base import SamplerConfig
from .combined import CombinedSampler


class TestCombined(unittest.TestCase):
    """Test cases for combined sampler."""

    @staticmethod
    def _get_boxes_targets(
        num_gts: int, num_samples: int
    ) -> Tuple[List[MatchResult], List[Boxes2D], List[Boxes2D]]:
        """Generate match, box target."""
        state = torch.random.get_rng_state()
        torch.random.set_rng_state(torch.manual_seed(0).get_state())
        matching = [
            MatchResult(
                assigned_gt_indices=torch.randint(0, num_gts, (num_samples,)),
                assigned_gt_iou=torch.rand(num_samples),
                assigned_labels=torch.randint(-1, 2, (num_samples,)),
            )
        ]
        boxes = [Boxes2D(torch.rand(num_samples, 5))]
        targets = [Boxes2D(torch.rand(num_gts, 5), torch.zeros(num_gts))]
        torch.random.set_rng_state(state)
        return matching, boxes, targets

    def test_sample(self) -> None:
        """Testcase for sample function."""
        samples_per_img = 256
        pos_fract = 0.5
        num_samples = 512
        num_gts = 3

        sampler = CombinedSampler(
            SamplerConfig(
                type="combined",
                batch_size_per_image=samples_per_img,
                positive_fraction=pos_fract,
                pos_strategy="instance_balanced",
                neg_strategy="iou_balanced",
            )
        )
        matching, boxes, targets = self._get_boxes_targets(
            num_gts, num_samples
        )
        sampled_boxes, sampled_targets = sampler.sample(
            matching, boxes, targets
        )
        self.assertEqual(len(sampled_boxes[0]), samples_per_img)
        self.assertEqual(len(sampled_boxes[0]), len(sampled_targets[0]))

        sampled_idx = []
        for sampled_target in sampled_targets[0]:  # type: ignore
            found = False
            for i, target in enumerate(targets[0]):  # type: ignore
                if torch.isclose(target.boxes, sampled_target.boxes).all():
                    sampled_idx.append(i)
                    found = True
            self.assertTrue(found)

        self.assertEqual(set(sampled_idx), set(range(num_gts)))

        sampler = CombinedSampler(
            SamplerConfig(
                type="combined",
                batch_size_per_image=samples_per_img,
                positive_fraction=pos_fract,
                pos_strategy="instance_balanced",
                neg_strategy="iou_balanced",
                floor_thr=0.1,
                num_bins=1,
            )
        )
        matching, boxes, targets = self._get_boxes_targets(
            num_gts, num_samples
        )
        sampler.sample(matching, boxes, targets)

        matching, boxes, targets = self._get_boxes_targets(num_gts, 128)
        sampler.sample(matching, boxes, targets)



"""mmdetection detector wrapper."""
from typing import Dict, List, Optional, Tuple

import torch
from mmcv.runner.checkpoint import load_checkpoint
from mmdet.models import TwoStageDetector, build_detector

from vist.struct import Boxes2D, Images, InputSample, LossesType, ModelOutput

from ..base import BaseModelConfig
from .base import BaseTwoStageDetector
from .mmdet_utils import (
    MMTwoStageDetectorConfig,
    _parse_losses,
    detections_from_mmdet,
    get_img_metas,
    get_mmdet_config,
    proposals_from_mmdet,
    proposals_to_mmdet,
    results_from_mmdet,
    targets_to_mmdet,
)

MMDET_MODEL_PREFIX = "https://download.openmmlab.com/mmdetection/v2.0/"


class MMTwoStageDetector(BaseTwoStageDetector):
    """mmdetection two-stage detector wrapper."""

    def __init__(self, cfg: BaseModelConfig):
        """Init."""
        super().__init__()
        self.cfg = MMTwoStageDetectorConfig(**cfg.dict())
        self.mm_cfg = get_mmdet_config(self.cfg)
        self.mm_detector = build_detector(self.mm_cfg)
        assert isinstance(self.mm_detector, TwoStageDetector)
        self.mm_detector.init_weights()
        self.mm_detector.train()
        if self.cfg.weights is not None:
            if self.cfg.weights.startswith("mmdet://"):
                self.cfg.weights = MMDET_MODEL_PREFIX + self.cfg.weights.strip(
                    "mmdet://"
                )
            load_checkpoint(self.mm_detector, self.cfg.weights)

        self.register_buffer(
            "pixel_mean",
            torch.tensor(self.cfg.pixel_mean).view(-1, 1, 1),
            False,
        )
        self.register_buffer(
            "pixel_std", torch.tensor(self.cfg.pixel_std).view(-1, 1, 1), False
        )

    @property
    def device(self) -> torch.device:
        """Get device where detect input should be moved to."""
        return self.pixel_mean.device

    def preprocess_image(self, batched_inputs: List[InputSample]) -> Images:
        """Batch, pad (standard stride=32) and normalize the input images."""
        images = Images.cat([inp.image for inp in batched_inputs], self.device)
        images.tensor = (images.tensor - self.pixel_mean) / self.pixel_std
        return images

    def forward_train(
        self,
        batch_inputs: List[List[InputSample]],
    ) -> LossesType:
        """Forward pass during training stage.

        Returns a dict of loss tensors.
        """
        inputs = [inp[0] for inp in batch_inputs]

        targets = []
        for x in inputs:
            assert x.boxes2d is not None
            targets.append(x.boxes2d.to(self.device))

        images = self.preprocess_image(inputs)
        image_metas = get_img_metas(images)
        gt_bboxes, gt_labels = targets_to_mmdet(targets)
        losses = self.mm_detector.forward_train(
            images.tensor,
            image_metas,
            gt_bboxes,
            gt_labels,
        )
        return _parse_losses(losses)

    def forward_test(
        self, batch_inputs: List[InputSample], postprocess: bool = True
    ) -> ModelOutput:
        """Forward pass during testing stage.

        Returns predictions for each input.
        """
        images = self.preprocess_image(batch_inputs)
        image_metas = get_img_metas(images)
        outs = self.mm_detector.simple_test(images.tensor, image_metas)
        detections = results_from_mmdet(outs, self.device)

        if postprocess:
            for inp, det in zip(batch_inputs, detections):
                ori_wh = (
                    batch_inputs[0].metadata.size.width,  # type: ignore
                    batch_inputs[0].metadata.size.height,  # type: ignore
                )
                self.postprocess(ori_wh, inp.image.image_sizes[0], det)

        return dict(detect=detections)  # type: ignore

    def extract_features(self, images: Images) -> Dict[str, torch.Tensor]:
        """Detector feature extraction stage.

        Return preprocessed images, backbone output features.
        """
        outs = self.mm_detector.extract_feat(images.tensor)
        if self.cfg.backbone_output_names is None:  # pragma: no cover
            return {f"out{i}": v for i, v in enumerate(outs)}

        return dict(zip(self.cfg.backbone_output_names, outs))

    def generate_proposals(
        self,
        images: Images,
        features: Dict[str, torch.Tensor],
        targets: Optional[List[Boxes2D]] = None,
    ) -> Tuple[List[Boxes2D], LossesType]:
        """Detector RPN stage.

        Return proposals per image and losses (empty if no targets).
        """
        feat_list = list(features.values())
        img_metas = get_img_metas(images)
        if targets is not None:
            gt_bboxes, _ = targets_to_mmdet(targets)

            proposal_cfg = self.mm_detector.train_cfg.get(
                "rpn_proposal", self.mm_detector.test_cfg.rpn
            )
            rpn_losses, proposals = self.mm_detector.rpn_head.forward_train(
                feat_list,
                img_metas,
                gt_bboxes,
                gt_labels=None,
                proposal_cfg=proposal_cfg,
            )
        else:
            proposals = self.mm_detector.rpn_head.simple_test(
                feat_list, img_metas
            )
            rpn_losses = dict()

        return proposals_from_mmdet(proposals), _parse_losses(rpn_losses)

    def generate_detections(
        self,
        images: Images,
        features: Dict[str, torch.Tensor],
        proposals: List[Boxes2D],
        targets: Optional[List[Boxes2D]] = None,
        compute_detections: bool = True,
    ) -> Tuple[Optional[List[Boxes2D]], LossesType]:
        """Detector second stage (RoI Head).

        Return losses (empty if no targets) and optionally detections.
        """
        proposal_list = proposals_to_mmdet(proposals)
        feat_list = list(features.values())
        img_metas = get_img_metas(images)
        if targets is not None:
            gt_bboxes, gt_labels = targets_to_mmdet(targets)
            detect_losses = self.mm_detector.roi_head.forward_train(
                feat_list,
                img_metas,
                proposal_list,
                gt_bboxes,
                gt_labels,
            )
            detect_losses = _parse_losses(detect_losses)
            assert (
                not compute_detections
            ), "mmdetection does not compute detections during train!"
            detections = None
        else:
            bboxes, labels = self.mm_detector.roi_head.simple_test_bboxes(
                feat_list,
                img_metas,
                proposal_list,
                self.mm_detector.roi_head.test_cfg,
            )
            detections = detections_from_mmdet(bboxes, labels)
            detect_losses = dict()

        return detections, detect_losses



"""Build VisT data loading pipeline."""
import itertools
import logging
from typing import Any, Dict, List, Optional, Union

import torch
from detectron2.config import CfgNode
from detectron2.data import build_batch_data_loader, detection_utils
from detectron2.data.catalog import DatasetCatalog, Metadata, MetadataCatalog
from detectron2.data.common import DatasetFromList
from detectron2.data.samplers import (
    InferenceSampler,
    RepeatFactorTrainingSampler,
    TrainingSampler,
)
from pydantic import BaseModel
from scalabel.label.typing import Frame

from .dataset_mapper import DataloaderConfig, DatasetMapper, MapDataset
from .samplers import TrackingInferenceSampler
from .utils import (
    discard_labels_outside_set,
    identity_batch_collator,
    prepare_labels,
    print_class_histogram,
)


class DataOptions(BaseModel):
    """Options for building the dataloader."""

    dataset: List[Dict[str, Any]]  # type: ignore
    sampler: Optional[Union[TrainingSampler, RepeatFactorTrainingSampler]]
    mapper: DatasetMapper
    total_batch_size: int
    num_workers: int

    class Config:  # needed due to sampler / mapper
        """Pydantic configuration for this particular class."""

        arbitrary_types_allowed = True


def get_dataset_frames(
    names: Union[str, List[str]],
    global_instance_ids: bool = False,
    categories: Optional[List[str]] = None,
) -> List[Frame]:
    """Load and prepare datasets in Scalabel format."""
    logger = logging.getLogger(__name__)
    if isinstance(names, str):
        names = [names]
    assert len(names), names
    dataset_frames = [
        DatasetCatalog.get(dataset_name) for dataset_name in names
    ]
    dataset_frames = list(itertools.chain.from_iterable(dataset_frames))

    has_instances = False
    for f in dataset_frames:
        if f.labels is not None:
            has_instances = True
            break

    if has_instances:
        if categories is not None:
            logger.info(
                "Filtering categories among the following datasets: %s", names
            )
            logger.info("Given categories: %s", categories)
            # synchronize metadata thing_classes, sync idx_to_class_mapping
            metas = [MetadataCatalog.get(d) for d in names]
            all_categories = list(
                itertools.chain.from_iterable(
                    [meta.thing_classes for meta in metas]
                )
            )
            # get intersection of classes among all datasets
            discard_set = [
                cat for cat in all_categories if cat not in categories
            ]
            # log classes and discarded ones
            if len(discard_set) > 0:
                logger.info(
                    "Discarding the following categories: %s", discard_set
                )
            assert (
                len(categories) > 0
            ), f"Classes of datasets {names} have no intersection!"
            for name, meta in zip(names, metas):
                MetadataCatalog.pop(name)
                meta_dict = meta.as_dict()
                meta_dict.update(
                    dict(
                        thing_classes=categories,
                        idx_to_class_mapping=dict(enumerate(categories)),
                    )
                )
                MetadataCatalog[name] = Metadata(**meta_dict)

            discard_labels_outside_set(dataset_frames, categories)

        # check metadata consistency
        detection_utils.check_metadata_consistency("thing_classes", names)

        # add category and instance ids, print class frequencies
        cat_name2id = {
            v: k
            for k, v in MetadataCatalog.get(
                names[0]
            ).idx_to_class_mapping.items()
        }
        frequencies = prepare_labels(
            cat_name2id, dataset_frames, global_instance_ids
        )
        print_class_histogram(frequencies)

    elif categories is not None:
        metas = [MetadataCatalog.get(d) for d in names]
        for name, meta in zip(names, metas):
            MetadataCatalog.pop(name)
            meta_dict = meta.as_dict()
            meta_dict.update(
                dict(
                    thing_classes=categories,
                    idx_to_class_mapping=dict(enumerate(categories)),
                )
            )
            MetadataCatalog[name] = Metadata(**meta_dict)

    assert len(dataset_frames) > 0, "No valid data found in {}.".format(
        ",".join(names)
    )
    return dataset_frames


def _train_loader_from_config(
    loader_cfg: DataloaderConfig, cfg: CfgNode
) -> DataOptions:
    """Construct training data loader from config."""
    dataset = get_dataset_frames(
        cfg.DATASETS.TRAIN,
        loader_cfg.compute_global_instance_ids,
        loader_cfg.categories,
    )
    mapper = DatasetMapper(loader_cfg)

    sampler_name = cfg.DATALOADER.SAMPLER_TRAIN
    if sampler_name == "TrainingSampler":
        sampler = TrainingSampler(len(dataset))
    else:
        raise ValueError("Unknown training sampler: {}".format(sampler_name))

    return DataOptions(
        dataset=dataset,
        sampler=sampler,
        mapper=mapper,
        total_batch_size=cfg.SOLVER.IMS_PER_BATCH,
        num_workers=loader_cfg.workers_per_gpu,
    )


def build_train_loader(
    loader_cfg: DataloaderConfig, det2cfg: CfgNode
) -> torch.utils.data.DataLoader:
    """Build train dataloader with some default features."""
    data_options = _train_loader_from_config(loader_cfg, det2cfg)
    dataset = DatasetFromList(data_options.dataset, copy=False)
    dataset = MapDataset(
        loader_cfg.ref_sampling_cfg, True, dataset, data_options.mapper
    )
    assert isinstance(data_options.sampler, torch.utils.data.sampler.Sampler)
    # aspect_ratio_grouping: tracking datasets usually do not contain
    # sequences with vastly different aspect ratios --> set to False
    return build_batch_data_loader(
        dataset,
        data_options.sampler,
        data_options.total_batch_size,
        aspect_ratio_grouping=False,
        num_workers=data_options.num_workers,
    )


def _test_loader_from_config(
    loader_cfg: DataloaderConfig, dataset_name: str
) -> DataOptions:
    """Construct testing data loader from config."""
    dataset = get_dataset_frames(dataset_name, False, loader_cfg.categories)
    mapper = DatasetMapper(loader_cfg, is_train=False)

    return DataOptions(
        dataset=dataset,
        mapper=mapper,
        total_batch_size=1,
        num_workers=loader_cfg.workers_per_gpu,
    )


def build_test_loader(
    loader_cfg: DataloaderConfig,
    dataset_name: str,
    inference_sampling: str,
) -> torch.utils.data.DataLoader:
    """Build test dataloader with some default features."""
    data_options = _test_loader_from_config(loader_cfg, dataset_name)
    dataset = DatasetFromList(data_options.dataset, copy=False)
    dataset = MapDataset(
        loader_cfg.ref_sampling_cfg, False, dataset, data_options.mapper
    )
    sampler = (
        TrackingInferenceSampler(dataset)
        if inference_sampling == "sequence_based"
        else InferenceSampler(len(dataset))
    )
    batch_sampler = torch.utils.data.sampler.BatchSampler(
        sampler, 1, drop_last=False
    )
    return torch.utils.data.DataLoader(
        dataset,
        num_workers=data_options.num_workers,
        batch_sampler=batch_sampler,
        collate_fn=identity_batch_collator,
    )



"""Init sampler module."""
from .base import BaseRoIPooler, RoIPoolerConfig, build_roi_pooler
from .d2_roi_pooler import D2RoIPooler

__all__ = [
    "BaseRoIPooler",
    "D2RoIPooler",
    "build_roi_pooler",
    "RoIPoolerConfig",
]



"""Test cases for default boilerplate logic."""
import unittest

from .defaults import default_argument_parser


class TestDefaults(unittest.TestCase):
    """Test cases for default boilerplate logic."""

    def test_argparse(self) -> None:
        """Check cmd line argument parsing."""
        parser = default_argument_parser()
        action = "train"
        cfg_path = "/path/to/config.toml"
        args = parser.parse_args([action, "--config", cfg_path, "--visualize"])
        self.assertEqual(args.config, cfg_path)
        self.assertEqual(args.action, action)
        self.assertEqual(args.visualize, True)



"""Dataset loader for scalabel format."""

from scalabel.label.io import load, load_label_config
from scalabel.label.typing import Dataset

from .base import BaseDatasetLoader


class Scalabel(BaseDatasetLoader):
    """Scalabel dataloading class."""

    def load_dataset(self) -> Dataset:
        """Load Scalabel frames from json."""
        assert self.cfg.annotations is not None
        dataset = load(
            self.cfg.annotations,
            validate_frames=self.cfg.validate_frames,
            nprocs=self.cfg.num_processes,
        )
        metadata_cfg = dataset.config
        if self.cfg.config_path is not None:
            metadata_cfg = load_label_config(self.cfg.config_path)
        assert metadata_cfg is not None
        dataset.config = metadata_cfg
        return dataset



"""Standard backend for local files on a hard drive."""
import os

from .base import BaseDataBackend, DataBackendConfig


class FileBackend(BaseDataBackend):
    """Raw file from hard disk data backend."""

    def __init__(self, cfg: DataBackendConfig):
        """Init."""
        super().__init__()
        self.cfg = DataBackendConfig(**cfg.dict())

    def get(self, filepath: str) -> bytes:
        """Get file content as bytes."""
        if not os.path.exists(filepath):
            raise FileNotFoundError(f"File not found:" f" {filepath}")
        with open(filepath, "rb") as f:
            value_buf = f.read()
        return value_buf



"""Base classes for data structures in VisT."""
import abc
from typing import Any, Dict, List, Optional, Union

import numpy as np
import numpy.typing as npt
import torch
from scalabel.eval.mot import EvalResults as MOTEvalResults
from scalabel.label.typing import Label


class DataInstance(metaclass=abc.ABCMeta):
    """Meta class for input data."""

    @abc.abstractmethod
    def to(  # pylint: disable=invalid-name
        self, device: torch.device
    ) -> "DataInstance":
        """Move to device (CPU / GPU / ...)."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def device(self) -> torch.device:
        """Returns current device if applicable."""
        raise NotImplementedError


class LabelInstance(DataInstance, metaclass=abc.ABCMeta):
    """Interface for bounding boxes, masks etc."""

    @classmethod
    @abc.abstractmethod
    def from_scalabel(
        cls,
        labels: List[Label],
        class_to_idx: Dict[str, int],
        label_id_to_idx: Optional[Dict[str, int]] = None,
    ) -> "LabelInstance":
        """Convert from scalabel format to ours."""
        raise NotImplementedError

    @abc.abstractmethod
    def to_scalabel(self, idx_to_class: Dict[int, str]) -> List[Label]:
        """Convert from ours to scalabel format."""
        raise NotImplementedError


NDArrayF64 = npt.NDArray[np.float64]
NDArrayUI8 = npt.NDArray[np.uint8]
TorchCheckpoint = Dict[str, Union[int, str, Dict[str, NDArrayF64]]]
LossesType = Dict[str, torch.Tensor]
EvalResult = Union[Dict[str, float], MOTEvalResults]
EvalResults = Dict[str, Union[Dict[str, float], MOTEvalResults]]
ModelOutput = Dict[str, List[LabelInstance]]
DictStrAny = Dict[str, Any]  # type: ignore



"""Dataset loader for coco format."""
import json
import os

from scalabel.label.from_coco import coco_to_scalabel
from scalabel.label.io import load_label_config
from scalabel.label.typing import Dataset

from .base import BaseDatasetLoader


class COCO(BaseDatasetLoader):
    """COCO dataloading class."""

    def load_dataset(self) -> Dataset:
        """Convert COCO annotations to scalabel format and prepare them."""
        assert self.cfg.annotations is not None
        if not os.path.exists(self.cfg.annotations) or not os.path.isfile(
            self.cfg.annotations
        ):
            raise FileNotFoundError(
                f"COCO json file not found: {self.cfg.annotations}"
            )
        with open(self.cfg.annotations, "r") as f:
            coco_anns = json.load(f)
        frames, metadata_cfg = coco_to_scalabel(coco_anns)
        if self.cfg.config_path is not None:
            metadata_cfg = load_label_config(self.cfg.config_path)

        return Dataset(frames=frames, config=metadata_cfg)



"""Detector module."""
from .base import BaseDetector, BaseTwoStageDetector
from .d2_wrapper import D2TwoStageDetector
from .mmdet_wrapper import MMTwoStageDetector

__all__ = [
    "BaseDetector",
    "BaseTwoStageDetector",
    "D2TwoStageDetector",
    "MMTwoStageDetector",
]



"""Function for registering the datasets in VisT."""
import abc
import logging
import os
from typing import List, Optional

from detectron2.data.catalog import DatasetCatalog, MetadataCatalog
from fvcore.common.timer import Timer
from pydantic import BaseModel, validator
from scalabel.label.typing import Config as MetadataConfig
from scalabel.label.typing import Dataset, Frame
from scalabel.label.utils import get_leaf_categories

from vist.common.registry import RegistryHolder

logger = logging.getLogger(__name__)


class BaseDatasetConfig(BaseModel, extra="allow"):
    """Config for training/evaluation datasets."""

    name: str
    type: str
    data_root: str
    annotations: Optional[str]
    config_path: Optional[str]
    eval_metrics: List[str] = []
    inference_sampling: str = "sample_based"
    validate_frames: bool = False
    num_processes: int = 4

    @validator("inference_sampling", check_fields=False)
    def validate_inference_sampling(  # pylint: disable=no-self-argument,no-self-use,line-too-long
        cls, value: str
    ) -> str:
        """Check inference_sampling attribute."""
        if value not in ["sample_based", "sequence_based"]:
            raise ValueError(
                "inference_sampling must be sample_based or sequence_based"
            )
        return value


class BaseDatasetLoader(metaclass=RegistryHolder):
    """Interface for loading dataset to scalabel format."""

    def __init__(self, cfg: BaseDatasetConfig):
        """Init dataset loader."""
        super().__init__()
        self.cfg = cfg

    @abc.abstractmethod
    def load_dataset(self) -> Dataset:
        """Load and possibly convert dataset to scalabel format."""
        raise NotImplementedError


def load_dataset(dataset_cfg: BaseDatasetConfig) -> List[Frame]:
    """Load a dataset into Scalabel format frames given its config."""
    timer = Timer()
    dataset = build_dataset_loader(dataset_cfg).load_dataset()
    assert dataset.config is not None
    add_data_path(dataset_cfg.data_root, dataset.frames)
    add_metadata(dataset.config, dataset_cfg)
    logger.info(
        "Loading %s takes %s seconds.",
        dataset_cfg.name,
        "{:.2f}".format(timer.seconds()),
    )
    return dataset.frames


def build_dataset_loader(cfg: BaseDatasetConfig) -> BaseDatasetLoader:
    """Build a dataset loader."""
    registry = RegistryHolder.get_registry(BaseDatasetLoader)
    if cfg.type in registry:
        dataset_loader = registry[cfg.type](cfg)
        assert isinstance(dataset_loader, BaseDatasetLoader)
        return dataset_loader
    raise NotImplementedError(f"Dataset type {cfg.type} not found.")


def register_dataset(dataset_cfg: BaseDatasetConfig) -> None:
    """Register a dataset."""
    DatasetCatalog.register(
        dataset_cfg.name, lambda: load_dataset(dataset_cfg)
    )


def add_data_path(data_root: str, frames: List[Frame]) -> None:
    """Add filepath to frame using data_root and frame.name."""
    for ann in frames:
        assert ann.name is not None
        if ann.video_name is not None:
            ann.url = os.path.join(data_root, ann.video_name, ann.name)
        else:
            ann.url = os.path.join(data_root, ann.name)


def add_metadata(
    metadata_cfg: MetadataConfig, dataset_cfg: BaseDatasetConfig
) -> None:
    """Add metadata to MetadataCatalog."""
    meta = MetadataCatalog.get(dataset_cfg.name)
    if meta.get("thing_classes") is None:
        cat_name2id = {
            cat.name: i + 1
            for i, cat in enumerate(
                get_leaf_categories(metadata_cfg.categories)
            )
        }
        meta.thing_classes = list(cat_name2id.keys())
        meta.idx_to_class_mapping = {v: k for k, v in cat_name2id.items()}
        meta.metadata_cfg = metadata_cfg
        meta.annotations = dataset_cfg.annotations
        meta.data_root = dataset_cfg.data_root
        meta.cfg_path = dataset_cfg.config_path



"""Init io module."""
from .base import BaseDataBackend, DataBackendConfig, build_data_backend
from .file import FileBackend
from .hdf5 import HDF5Backend

__all__ = [
    "HDF5Backend",
    "FileBackend",
    "build_data_backend",
    "DataBackendConfig",
]



"""Load and convert bdd100k labels to scalabel format."""
import inspect
import os

from scalabel.label.io import load, load_label_config, save
from scalabel.label.typing import Dataset

from .base import BaseDatasetConfig, BaseDatasetLoader

try:
    from scalabel.label.from_waymo import from_waymo

    WAYMO_INSTALLED = True  # pragma: no cover
except NameError:  # pragma: no cover
    WAYMO_INSTALLED = False  # pragma: no cover


class WaymoDatasetConfig(BaseDatasetConfig):
    """Config for training/evaluation datasets."""

    input_dir: str
    output_dir: str
    save_images: bool = True
    use_lidar_labels: bool = False


class Waymo(BaseDatasetLoader):  # pragma: no cover
    """Waymo Open dataloading class."""

    def __init__(self, cfg: BaseDatasetConfig):
        """Init dataset loader."""
        super().__init__(cfg)
        self.cfg = WaymoDatasetConfig(**cfg.dict())  # type: WaymoDatasetConfig

    def load_dataset(self) -> Dataset:
        """Convert Waymo annotations to Scalabel format."""
        assert (
            WAYMO_INSTALLED
        ), "Using waymo dataset needs waymo open dataset reader installed!."
        assert (
            self.cfg.data_root == self.cfg.output_dir
        ), "Waymo requires conversion output path to be equal to data_root."
        cfg_path = self.cfg.config_path
        if cfg_path is None:
            cfg_path = os.path.join(
                os.path.dirname(os.path.abspath(inspect.stack()[1][1])),
                "waymo.toml",
            )
        metadata_cfg = load_label_config(cfg_path)

        scalabel_anns_path = os.path.join(
            self.cfg.output_dir, "scalabel_anns.json"
        )
        if not os.path.exists(scalabel_anns_path):
            frames = from_waymo(
                self.cfg.input_dir,
                self.cfg.output_dir,
                self.cfg.save_images,
                self.cfg.use_lidar_labels,
                self.cfg.num_processes,
            )
            save(scalabel_anns_path, frames)
        else:
            frames = load(
                scalabel_anns_path,
                validate_frames=self.cfg.validate_frames,
                nprocs=self.cfg.num_processes,
            ).frames
        return Dataset(frames=frames, config=metadata_cfg)



"""Utility functions for bounding boxes."""
import torch
from detectron2.structures import Boxes, pairwise_iou

from vist.struct import Boxes2D


def compute_iou(boxes1: Boxes2D, boxes2: Boxes2D) -> torch.Tensor:
    """Compute IoU between all pairs of boxes.

    Args:
        boxes1, boxes2 (Boxes2D): Contains N & M boxes.

    Returns:
        Tensor: IoU, size [N, M].
    """
    return pairwise_iou(Boxes(boxes1.boxes[:, :4]), Boxes(boxes2.boxes[:, :4]))


def random_choice(tensor: torch.Tensor, sample_size: int) -> torch.Tensor:
    """Randomly choose elements from a tensor."""
    perm = torch.randperm(len(tensor), device=tensor.device)[:sample_size]
    return tensor[perm]


def non_intersection(t1: torch.Tensor, t2: torch.Tensor) -> torch.Tensor:
    """Get the elements of t1 that are not present in t2."""
    compareview = t2.repeat(t1.shape[0], 1).T
    return t1[(compareview != t1).T.prod(1) == 1]



"""Utilities for visualization."""
import colorsys
from typing import List, Tuple, Union

import numpy as np
import torch
from PIL import Image
from scipy.spatial.transform import Rotation as R

from vist.struct import Boxes2D, Boxes3D, NDArrayF64, NDArrayUI8

ImageType = Union[torch.Tensor, NDArrayUI8, NDArrayF64]

BoxType = Union[Boxes2D, List[Boxes2D]]
Box3DType = Union[Boxes3D, List[Boxes3D]]

ColorType = Union[
    Union[Tuple[int], str],
    List[Union[Tuple[int], str]],
    List[List[Union[Tuple[int], str]]],
]


def generate_colors(length: int) -> List[Tuple[int]]:
    """Generate a color palette of [length] colors."""
    brightness = 0.7
    hsv = [(i / length, 1, brightness) for i in range(length)]
    colors = list(map(lambda c: colorsys.hsv_to_rgb(*c), hsv))
    colors = (np.array(colors) * 255).astype(np.uint8).tolist()  # type: ignore
    s = np.random.get_state()
    np.random.seed(0)
    result = [tuple(colors[i]) for i in np.random.permutation(len(colors))]
    np.random.set_state(s)
    return result  # type: ignore


NUM_COLORS = 50
COLOR_PALETTE = generate_colors(NUM_COLORS)


def preprocess_boxes(
    boxes: Union[BoxType, Box3DType], color_idx: int = 0
) -> Tuple[List[List[float]], List[Tuple[int]], List[str]]:
    """Preprocess BoxType to boxes / colors / labels for drawing."""
    if isinstance(boxes, list):
        result_box, result_color, result_labels = [], [], []
        for i, b in enumerate(boxes):
            box, color, labels = preprocess_boxes(b, i)  # type: ignore
            result_box.extend(box)
            result_color.extend(color)
            result_labels.extend(labels)
        return result_box, result_color, result_labels

    assert isinstance(boxes, (Boxes2D, Boxes3D))

    if boxes.boxes.shape[-1] in [5, 8, 10]:  # 4, 7, 9 DoF boxes + score
        boxes_list = boxes.boxes[:, :-1].cpu().numpy().tolist()
        scores = boxes.boxes[:, -1].cpu().numpy().tolist()
    else:
        boxes_list = boxes.boxes.cpu().numpy().tolist()
        scores = [None for _ in range(len(boxes_list))]

    if boxes.track_ids is not None:
        track_ids = boxes.track_ids.cpu().numpy()
        if len(track_ids.shape) > 1:
            track_ids = track_ids.squeeze(-1)
    else:
        track_ids = [None for _ in range(len(boxes_list))]

    if boxes.class_ids is not None:
        class_ids = boxes.class_ids.cpu().numpy()
    else:
        class_ids = [None for _ in range(len(boxes_list))]

    labels, draw_colors = [], []
    for s, t, c in zip(scores, track_ids, class_ids):
        if t is not None:
            draw_color = COLOR_PALETTE[int(t) % NUM_COLORS]
        elif c is not None:
            draw_color = COLOR_PALETTE[int(c) % NUM_COLORS]
        else:
            draw_color = COLOR_PALETTE[color_idx % NUM_COLORS]

        label = ""
        if t is not None:
            label += str(int(t)) + ","
        if c is not None:
            str_c = str(int(c))
            if boxes.metadata is not None:
                str_c = boxes.metadata[str_c]  # type: ignore
            label += str_c + ","

        if s is not None:
            label += "{:.1f}%".format(s * 100)
        labels.append(label)
        draw_colors.append(draw_color)

    return boxes_list, draw_colors, labels


def preprocess_image(image: ImageType, mode: str = "BGR") -> Image.Image:
    """Validate and convert input image.

    Args:
        image: CHW or HWC image (ImageType) with C = 3.
        mode: input channel format (e.g. BGR, HSV). More info
        at https://pillow.readthedocs.io/en/stable/handbook/concepts.html

    Returns:
        PIL.Image.Image: Processed image in RGB.
    """
    assert len(image.shape) == 3
    assert image.shape[0] == 3 or image.shape[-1] == 3

    if isinstance(image, torch.Tensor):
        image = image.cpu().numpy()

    if not image.shape[-1] == 3:
        image = image.transpose(1, 2, 0)
    min_val, max_val = (
        np.min(image, axis=(0, 1)),  # type: ignore
        np.max(image, axis=(0, 1)),  # type: ignore
    )

    image = image.astype(np.float32)

    image = (image - min_val) / (max_val - min_val) * 255.0

    if mode == "BGR":
        image = image[..., [2, 1, 0]]
        mode = "RGB"

    return Image.fromarray(image.astype(np.uint8), mode=mode).convert("RGB")


def box3d_to_corners(box3d: List[float]) -> NDArrayF64:
    """Convert Boxes3D style box to its respective corner points."""
    x_loc, y_loc, z_loc = box3d[:3]
    h, w, l = box3d[3:6]
    if len(box3d) < 9:  # if 7 DoF box type
        rx, ry, rz = 0.0, box3d[6], 0.0
    else:
        rx, ry, rz = box3d[6], box3d[7], box3d[8]

    x_corners = np.array(
        [
            l / 2.0,
            l / 2.0,
            -l / 2.0,
            -l / 2.0,
            l / 2.0,
            l / 2.0,
            -l / 2.0,
            -l / 2.0,
        ],
        dtype=np.float64,
    )
    z_corners = np.array(
        [
            w / 2.0,
            -w / 2.0,
            -w / 2.0,
            w / 2.0,
            w / 2.0,
            -w / 2.0,
            -w / 2.0,
            w / 2.0,
        ],
        dtype=np.float64,
    )
    y_corners = np.zeros((8,), dtype=np.float64)
    y_corners[0:4] = h / 2
    y_corners[4:8] = -h / 2

    rot = R.from_euler("xyz", np.array([rx, ry, rz])).as_matrix()
    temp_corners = np.concatenate(
        (
            x_corners.reshape(8, 1),
            y_corners.reshape(8, 1),
            z_corners.reshape(8, 1),
        ),
        axis=1,
    )  # type: ignore
    corners = np.matmul(temp_corners, rot.T)
    corners[:, 0], corners[:, 1], corners[:, 2] = (
        corners[:, 0] + x_loc,
        corners[:, 1] + y_loc,
        corners[:, 2] + z_loc,
    )
    return corners  # type: ignore



"""Interface for VisT bounding box samplers."""

import abc
from typing import List, Tuple

from pydantic import BaseModel, Field

from vist.common.registry import RegistryHolder
from vist.struct import Boxes2D

from ..matchers.base import MatchResult


class SamplerConfig(BaseModel, extra="allow"):
    """Sampler base config."""

    # Field(...) necessary for linter
    # See https://github.com/samuelcolvin/pydantic/issues/1899
    type: str = Field(...)
    batch_size_per_image: int = Field(...)
    positive_fraction: float = Field(...)


class BaseSampler(metaclass=RegistryHolder):
    """Sampler base class."""

    @abc.abstractmethod
    def sample(
        self,
        matching: List[MatchResult],
        boxes: List[Boxes2D],
        targets: List[Boxes2D],
    ) -> Tuple[List[Boxes2D], List[Boxes2D]]:
        """Sample bounding boxes according to their struct."""
        raise NotImplementedError


def build_sampler(cfg: SamplerConfig) -> BaseSampler:
    """Build a bounding box sampler from config."""
    registry = RegistryHolder.get_registry(BaseSampler)
    if cfg.type in registry:
        module = registry[cfg.type](cfg)
        assert isinstance(module, BaseSampler)
        return module
    raise NotImplementedError(f"Sampler {cfg.type} not found.")



"""VisT Visualization tools for analysis and debugging."""
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from PIL import Image, ImageDraw, ImageFont
from scalabel.label.utils import project_points_to_image

from vist.struct import Intrinsics, NDArrayF64

from .utils import (
    Box3DType,
    BoxType,
    ImageType,
    box3d_to_corners,
    preprocess_boxes,
    preprocess_image,
)


def imshow(
    image: Union[Image.Image, ImageType], mode: str = "BGR"
) -> None:  # pragma: no cover
    """Imshow method.

    Args:
        image: PIL Image or ImageType (i.e. numpy array, torch.Tensor)
        mode: Image channel format, will be used to convert ImageType to
        an RGB PIL Image. Not necessary if 'image' is an RGB PIL Image.
    """
    if not isinstance(image, Image.Image):
        image = preprocess_image(image, mode)
    plt.imshow(np.asarray(image))
    plt.show()


def imshow_bboxes(
    image: ImageType, boxes: BoxType, mode: str = "BGR"
) -> None:  # pragma: no cover
    """Show image with bounding boxes."""
    image = preprocess_image(image, mode)
    box_list, color_list, label_list = preprocess_boxes(boxes)
    for box, col, label in zip(box_list, color_list, label_list):
        draw_bbox(image, box, col, label)

    imshow(image)


def imshow_bboxes3d(
    image: ImageType,
    boxes: Box3DType,
    intrinsics: Union[NDArrayF64, Intrinsics],
    mode: str = "BGR",
) -> None:  # pragma: no cover
    """Show image with bounding boxes."""
    image = preprocess_image(image, mode)
    box_list, color_list, label_list = preprocess_boxes(boxes)
    if isinstance(intrinsics, Intrinsics):
        intrinsic_matrix = intrinsics.tensor.cpu().numpy()  # type: NDArrayF64
    elif isinstance(intrinsics, np.ndarray):
        intrinsic_matrix = intrinsics
    else:
        raise ValueError(f"Invalid type for intrinsics: {type(intrinsics)}")

    assert intrinsic_matrix.shape == (
        3,
        3,
    ), f"Intrinsics must be of shape 3x3, got {intrinsic_matrix.shape}"

    for box, col, label in zip(box_list, color_list, label_list):
        draw_bbox3d(image, box, intrinsic_matrix, col, label)

    imshow(image)


def draw_bbox(
    image: Image.Image,
    box: List[float],
    color: Tuple[int],
    label: Optional[str] = None,
) -> None:
    """Draw 2D box onto image."""
    draw = ImageDraw.Draw(image)
    draw.rectangle(box, outline=color)
    if label is not None:
        font = ImageFont.load_default()
        draw.text(box[:2], label, (255, 255, 255), font=font)


def draw_bbox3d(
    image: Image.Image,
    box3d: List[float],
    intrinsics: NDArrayF64,
    color: Tuple[int],
    label: Optional[str] = None,
) -> None:  # pragma: no cover
    """Draw 3D box onto image."""
    draw = ImageDraw.Draw(image)
    corners = project_points_to_image(box3d_to_corners(box3d), intrinsics)

    def draw_rect(selected_corners: NDArrayF64) -> None:
        prev = selected_corners[-1]
        for corner in selected_corners:
            draw.line((tuple(prev), tuple(corner)), fill=color)
            prev = corner

    # Draw the sides
    for i in range(4):
        draw.line((tuple(corners[i]), tuple(corners[i + 4])), fill=color)

    # Draw bottom (first 4 corners) and top (last 4 corners)
    draw_rect(corners[:4])
    draw_rect(corners[4:])

    # Draw line indicating the front
    center_bottom_forward = np.mean(corners[:2], axis=0)
    center_bottom = np.mean(corners[:4], axis=0)
    draw.line((tuple(center_bottom), tuple(center_bottom_forward)), fill=color)

    if label is not None:
        font = ImageFont.load_default()
        center_top_forward = tuple(np.mean(corners[2:4], axis=0))
        draw.text(center_top_forward, label, (255, 255, 255), font=font)



"""Testcases for data backends."""
import sys
import unittest

from vist.data.utils import str_decode

from .base import DataBackendConfig
from .file import FileBackend
from .hdf5 import HDF5Backend


class TestBackends(unittest.TestCase):
    """Testcases for ensuring equal output of each backend."""

    def test_get(self) -> None:
        """Test image retrieval from different backends."""
        backend_file = FileBackend(DataBackendConfig(type="file"))
        backend_hdf5 = HDF5Backend(DataBackendConfig(type="hdf5"))
        sample_path = (
            "vist/engine/testcases/track/bdd100k-samples/images/"
            "00091078-875c1f73/00091078-875c1f73-0000166.jpg"
        )
        hdf5_path = (
            "vist/engine/testcases/track/bdd100k-samples/images.hdf5/"
            "00091078-875c1f73/00091078-875c1f73-0000166.jpg"
        )

        out_file = backend_file.get(sample_path)
        out_hdf5 = backend_hdf5.get(hdf5_path)
        self.assertTrue(out_file == out_hdf5)

        # check db_cache
        backend_hdf5.get(hdf5_path)

        self.assertRaises(FileNotFoundError, backend_file.get, "invalid_path")
        self.assertRaises(FileNotFoundError, backend_hdf5.get, "invalid_path")

        invalid_hdf5_path = (
            "vist/engine/testcases/track/bdd100k-samples/images.hdf5/"
            "000/000.jpg"
        )
        self.assertRaises(ValueError, backend_hdf5.get, invalid_hdf5_path)

    def test_str_decode(self) -> None:
        """Test str decode method in utils."""
        my_str = "Hello world!"
        gen_str = str_decode(my_str.encode(sys.getdefaultencoding()))
        self.assertTrue(my_str == gen_str)



# type: ignore
"""Test cases for VisT engine Trainer."""

import shutil
import unittest
from argparse import Namespace

import torch

from vist import config
from vist.engine.trainer import predict
from vist.engine.trainer import test as evaluate
from vist.engine.trainer import train
from vist.unittest.utils import d2_data_reset, get_test_file

from .utils import _register


class BaseEngineTests:
    """Base class for engine tests."""

    class TestDetect(unittest.TestCase):
        """Test cases for vist models."""

        cfg = None
        work_dir = None
        args = None

        def test_predict(self) -> None:
            """Testcase for predict."""
            self.assertIsNotNone(self.cfg)
            self.cfg.launch.action = "predict"
            if torch.cuda.is_available():
                self.cfg.launch.device = "cuda"  # pragma: no cover
            self.cfg.launch.visualize = True
            predict(self.cfg)

        def test_train(self) -> None:
            """Testcase for training."""
            self.assertIsNotNone(self.cfg)
            self.cfg.launch.action = "train"
            self.cfg.launch.seed = 42
            if torch.cuda.is_available():
                self.cfg.launch.device = "cuda"  # pragma: no cover
            train(self.cfg)
            self.cfg.launch.seed = -1

        def test_testfunc(self) -> None:
            """Testcase for test function."""
            self.assertIsNotNone(self.cfg)
            self.cfg.launch.action = "predict"
            if torch.cuda.is_available():
                self.cfg.launch.device = "cuda"  # pragma: no cover
            results = evaluate(self.cfg)["detect"]
            metric_keys = [
                "AP",
                "AP_50",
                "AP_75",
                "AP_small",
                "AP_medium",
                "AP_large",
                "AR_max_1",
                "AR_max_10",
                "AR_max_100",
                "AR_small",
                "AR_medium",
                "AR_large",
                "AP_pedestrian",
                "AP_rider",
                "AP_car",
                "AP_truck",
                "AP_bus",
                "AP_train",
                "AP_motorcycle",
                "AP_bicycle",
                "AP_traffic light",
                "AP_traffic sign",
            ]

            for k in results:
                self.assertIn(k, metric_keys)

        @classmethod
        def tearDownClass(cls) -> None:
            """Clean up dataset registry, files."""
            assert cls.cfg.train is not None, "cfg.train must not be None"
            assert cls.cfg.test is not None, "cfg.test must not be None"
            d2_data_reset([ds.name for ds in cls.cfg.train])
            d2_data_reset([ds.name for ds in cls.cfg.test])
            d2_data_reset(["00091078-875c1f73"])
            shutil.rmtree(cls.work_dir, ignore_errors=True)

    class TestTrack(unittest.TestCase):
        """Test cases for vist tracking."""

        cfg = None
        work_dir = None
        args = None
        metric_keys = [
            "car",
            "pedestrian",
            "rider",
            "truck",
            "bus",
            "train",
            "motorcycle",
            "bicycle",
            "human",
            "vehicle",
            "bike",
            "OVERALL",
        ]
        predict_dir = (
            "vist/engine/testcases/track/bdd100k-samples/images/"
            "00091078-875c1f73/"
        )

        def test_predict(self) -> None:
            """Testcase for predict."""
            self.assertIsNotNone(self.cfg)
            self.cfg.launch.action = "predict"
            if torch.cuda.is_available():
                self.cfg.launch.device = "cuda"  # pragma: no cover
            self.cfg.launch.input_dir = self.predict_dir
            self.cfg.launch.visualize = True
            predict(self.cfg)

        def test_testfunc(self) -> None:
            """Testcase for test function."""
            self.assertIsNotNone(self.cfg)
            self.cfg.launch.action = "predict"
            if torch.cuda.is_available():
                self.cfg.launch.device = "cuda"  # pragma: no cover
            results = evaluate(self.cfg)["track"]

            for k in results:
                self.assertIn(k, self.metric_keys)

        def test_train(self) -> None:
            """Testcase for training."""
            self.assertIsNotNone(self.cfg)
            self.cfg.launch.action = "train"
            self.cfg.launch.seed = 42
            if torch.cuda.is_available():
                self.cfg.launch.device = "cuda"  # pragma: no cover
            train(self.cfg)
            self.cfg.launch.seed = -1

        def test_duplicate_register(self) -> None:
            """Test if duplicated datasets are skipped."""
            self.assertIsNotNone(self.cfg.train)
            self.assertIsNotNone(self.cfg.test)
            _register(self.cfg.train)
            _register(self.cfg.test)
            _register(self.cfg.train)

        @classmethod
        def tearDownClass(cls) -> None:
            """Clean up dataset registry, files."""
            assert cls.cfg.train is not None, "cfg.train must not be None"
            assert cls.cfg.test is not None, "cfg.test must not be None"
            d2_data_reset([ds.name for ds in cls.cfg.train])
            d2_data_reset([ds.name for ds in cls.cfg.test])
            d2_data_reset(["00091078-875c1f73"])
            shutil.rmtree(cls.work_dir, ignore_errors=True)


class TestTrackD2(BaseEngineTests.TestTrack):
    """Detectron2 tracking test cases."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up class."""
        cls.work_dir = "./unittests/unittest_track_d2/"
        cls.args = Namespace(
            config=get_test_file("track/qdtrack_d2.toml"),
            output_dir=cls.work_dir,
        )
        cls.cfg = config.parse_config(cls.args)


class TestTrackMM(BaseEngineTests.TestTrack):
    """MMDetection tracking test cases."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up class."""
        cls.work_dir = "./unittests/unittest_track_mm/"
        cls.args = Namespace(
            config=get_test_file("track/qdtrack_mmdet.toml"),
            output_dir=cls.work_dir,
        )
        cls.cfg = config.parse_config(cls.args)


class TestTrackMMKITTI(BaseEngineTests.TestTrack):
    """MMDetection tracking test cases on KITTI."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up class."""
        cls.work_dir = "./unittests/unittest_track_mm_kitti/"
        cls.args = Namespace(
            config=get_test_file("track/qdtrack_mmdet_kitti.toml"),
            output_dir=cls.work_dir,
        )
        cls.cfg = config.parse_config(cls.args)
        cls.metric_keys = [
            "pedestrian",
            "cyclist",
            "car",
            "truck",
            "tram",
            "misc",
            "OVERALL",
        ]
        cls.predict_dir = (
            "vist/engine/testcases/track/kitti-samples/"
            "tracking/training/image_02/0001/"
        )


class TestDetectD2(BaseEngineTests.TestDetect):
    """Detectron2 detection test cases."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up class."""
        cls.work_dir = "./unittests/unittest_detect_d2/"
        cls.args = Namespace(
            config=get_test_file("detect/faster_rcnn_d2.toml"),
            output_dir=cls.work_dir,
        )
        cls.cfg = config.parse_config(cls.args)


class TestDetectMM(BaseEngineTests.TestDetect):
    """MMDetection detection test cases."""

    @classmethod
    def setUpClass(cls) -> None:
        """Set up class."""
        cls.work_dir = "./unittests/unittest_detect_mm/"
        args = Namespace(
            config=get_test_file("detect/faster_rcnn_mmdet.toml"),
            output_dir=cls.work_dir,
        )
        cls.cfg = config.parse_config(args)



"""Datasets module."""
from .base import BaseDatasetConfig, BaseDatasetLoader, register_dataset
from .bdd100k import BDD100K
from .coco import COCO
from .custom import Custom
from .kitti import KITTI
from .motchallenge import MOTChallenge
from .scalabel import Scalabel
from .waymo import Waymo

__all__ = [
    "register_dataset",
    "BaseDatasetLoader",
    "BaseDatasetConfig",
    "BDD100K",
    "Scalabel",
    "COCO",
    "Custom",
    "MOTChallenge",
    "Waymo",
    "KITTI",
]



"""VisT augmentations."""
import random
from typing import List, Tuple, Union

import torch
import torch.nn.functional as F

from .base import AugParams, BaseAugmentation


class Resize(BaseAugmentation):
    """Simple resize augmentation class."""

    def __init__(
        self,
        shape: Union[Tuple[int, int], List[Tuple[int, int]]],
        keep_ratio: bool = False,
        multiscale_mode: str = "range",
        scale_range: Tuple[float, float] = (1.0, 1.0),
        interpolation: str = "bilinear",
    ) -> None:
        """Init function.

        Args:
            shape: Image shape to be resized to in (H, W) format. In
            multiscale mode 'list', shape represents the list of possible
            shapes for resizing.
            interpolation: Interpolation method. One of ["nearest", "bilinear",
            "bicubic"]
            keep_ratio: If aspect ratio of original image should be kept, the
            new height will be scaled according to the new width and the
            aspect ratio of the original image as:
            new_h = new_w / (orginal_w / original_h)
            multiscale_mode: one of [range, list],
            scale_range: Range of sampled image scales in range mode, e.g.
            (0.8, 1.2), indicating minimum of 0.8 * shape and maximum of
            1.2 * shape.
            return_transform: If the transform should be returned in matrix
            format.
        """
        super().__init__(p=1.0)
        self.shape = shape
        self.keep_ratio = keep_ratio
        self.multiscale_mode = multiscale_mode
        assert self.multiscale_mode in ["list", "range"]
        self.scale_range = scale_range
        if self.multiscale_mode == "list":
            assert isinstance(
                self.shape, list
            ), "Specify shape as list when using multiscale mode list."
            assert len(self.shape) >= 1
            self.shape = (
                [  # pylint: disable=unsubscriptable-object,line-too-long
                    (int(s[0]), int(s[1])) for s in self.shape
                ]
            )
        else:
            assert isinstance(
                self.shape, tuple
            ), "Specify shape as tuple when using multiscale mode range."
            self.shape = (int(self.shape[0]), int(self.shape[1]))
            assert (
                scale_range[0] <= scale_range[1]
            ), f"Invalid scale range: {scale_range[1]} < {scale_range[0]}"

        self.interpolation = interpolation
        assert self.interpolation in ["nearest", "bilinear", "bicubic"]

    def generate_parameters(self, batch_shape: torch.Size) -> AugParams:
        """Generate current parameters."""
        if self.multiscale_mode == "range":
            assert isinstance(self.shape, tuple)
            if self.scale_range[0] < self.scale_range[1]:  # do multi-scale
                w_scale = (
                    random.random()
                    * (self.scale_range[1] - self.scale_range[0])
                    + self.scale_range[0]
                )
                h_scale = (
                    random.random()
                    * (self.scale_range[1] - self.scale_range[0])
                    + self.scale_range[0]
                )
                shape = (
                    int(self.shape[0] * h_scale),
                    int(self.shape[1] * w_scale),
                )
            else:
                shape = self.shape
        else:
            assert isinstance(self.shape, list)
            shape = random.sample(self.shape, k=1)[0]

        if self.keep_ratio:
            _, _, h, w = batch_shape
            shape = (int(shape[1] / (w / h)), shape[1])

        return dict(shape=shape)

    def compute_transformation(
        self, inputs: torch.Tensor, params: AugParams
    ) -> torch.Tensor:
        """Compute transformation for resize."""
        transform = torch.eye(3, device=inputs.device)
        n, _, h, w = inputs.shape
        transform[0, 0] = params["shape"][1] / w
        transform[1, 1] = params["shape"][0] / h
        return torch.stack([transform for _ in range(n)], 0)

    def apply_transform(
        self, inputs: torch.Tensor, params: AugParams, transform: torch.Tensor
    ) -> torch.Tensor:
        """Apply resize."""
        align_corners = None if self.interpolation == "nearest" else False
        output = F.interpolate(
            inputs,
            params["shape"],
            mode=self.interpolation,
            align_corners=align_corners,
        )
        return output



"""Similarity Head definition for quasi-dense instance similarity learning."""
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
from detectron2.layers.batch_norm import get_norm
from detectron2.layers.wrappers import Conv2d
from torch import nn

from vist.common.bbox.matchers import MatcherConfig, build_matcher
from vist.common.bbox.poolers import RoIPoolerConfig, build_roi_pooler
from vist.common.bbox.samplers import SamplerConfig, build_sampler
from vist.struct import Boxes2D, Images

from .base import BaseSimilarityHead, SimilarityLearningConfig


class QDSimilarityHeadConfig(SimilarityLearningConfig):
    """Quasi-dense Similarity Head config."""

    num_classes: int
    in_dim: int
    num_convs: int
    conv_out_dim: int
    conv_has_bias: bool
    num_fcs: int
    fc_out_dim: int
    embedding_dim: int
    norm: str
    proposal_append_gt: bool
    in_features: List[str] = ["p2", "p3", "p4", "p5"]
    proposal_pooler: RoIPoolerConfig
    proposal_sampler: SamplerConfig
    proposal_matcher: MatcherConfig


class QDSimilarityHead(BaseSimilarityHead):
    """Instance embedding head for quasi-dense similarity learning."""

    def __init__(self, cfg: SimilarityLearningConfig) -> None:
        """Init."""
        super().__init__()
        self.cfg = QDSimilarityHeadConfig(**cfg.dict())

        self.sampler = build_sampler(self.cfg.proposal_sampler)
        self.matcher = build_matcher(self.cfg.proposal_matcher)
        self.roi_pooler = build_roi_pooler(self.cfg.proposal_pooler)

        self.convs, self.fcs, last_layer_dim = self._init_embedding_head()
        self.fc_embed = nn.Linear(last_layer_dim, self.cfg.embedding_dim)
        self._init_weights()

    def _init_weights(self) -> None:
        """Init weights of modules in head."""
        for m in self.convs:
            nn.init.kaiming_uniform_(m.weight, a=1)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)  # pragma: no cover

        for m in self.fcs:
            if isinstance(m[0], nn.Linear):
                nn.init.xavier_uniform_(m[0].weight)
                nn.init.constant_(m[0].bias, 0)

        nn.init.normal_(self.fc_embed.weight, 0, 0.01)
        nn.init.constant_(self.fc_embed.bias, 0)

    def _init_embedding_head(
        self,
    ) -> Tuple[torch.nn.ModuleList, torch.nn.ModuleList, int]:
        """Init modules of head."""
        last_layer_dim = self.cfg.in_dim
        # add branch specific conv layers
        convs = nn.ModuleList()
        if self.cfg.num_convs > 0:
            for i in range(self.cfg.num_convs):
                conv_in_dim = (
                    last_layer_dim if i == 0 else self.cfg.conv_out_dim
                )
                convs.append(
                    Conv2d(
                        conv_in_dim,
                        self.cfg.conv_out_dim,
                        kernel_size=3,
                        padding=1,
                        bias=self.cfg.conv_has_bias,
                        norm=get_norm(self.cfg.norm, self.cfg.conv_out_dim),
                        activation=nn.ReLU(inplace=True),
                    )
                )
            last_layer_dim = self.cfg.conv_out_dim

        fcs = nn.ModuleList()
        if self.cfg.num_fcs > 0:
            last_layer_dim *= np.prod(self.cfg.proposal_pooler.resolution)
            for i in range(self.cfg.num_fcs):
                fc_in_dim = last_layer_dim if i == 0 else self.cfg.fc_out_dim
                fcs.append(
                    nn.Sequential(
                        nn.Linear(fc_in_dim, self.cfg.fc_out_dim),
                        nn.ReLU(inplace=True),
                    )
                )
            last_layer_dim = self.cfg.fc_out_dim
        return convs, fcs, last_layer_dim

    @torch.no_grad()  # type: ignore
    def match_and_sample_proposals(
        self, proposals: List[Boxes2D], targets: List[Boxes2D]
    ) -> Tuple[List[Boxes2D], List[Boxes2D]]:
        """Match proposals to targets and subsample."""
        if self.cfg.proposal_append_gt:
            proposals = [
                Boxes2D.cat([p, t]) for p, t in zip(proposals, targets)
            ]
        matching = self.matcher.match(proposals, targets)
        return self.sampler.sample(matching, proposals, targets)

    def forward(  # type: ignore # pylint: disable=arguments-differ
        self,
        images: Images,
        features: Dict[str, torch.Tensor],
        proposals: List[Boxes2D],
        targets: Optional[List[Boxes2D]] = None,
        filter_negatives: bool = False,
    ) -> Tuple[Tuple[torch.Tensor], Optional[List[Boxes2D]]]:
        """Forward of embedding head.

        We do not return a loss here, since the matching loss needs to
        be computed across all frames. Here, we only run the forward pass
        per frame, as well as proposal sampling and target assignment.
        """
        del images
        features_list = [features[f] for f in self.cfg.in_features]
        if self.training:
            assert targets is not None, "targets required during training"
            proposals, targets = self.match_and_sample_proposals(
                proposals, targets
            )
            if filter_negatives:
                proposals = [
                    p[t.class_ids != -1] for p, t in zip(proposals, targets)  # type: ignore # pylint: disable=line-too-long
                ]
                targets = [t[t.class_ids != -1] for t in targets]  # type: ignore # pylint: disable=line-too-long

        x = self.roi_pooler.pool(features_list, proposals)
        # convs
        if self.cfg.num_convs > 0:
            for conv in self.convs:
                x = conv(x)

        # fcs
        x = torch.flatten(x, start_dim=1)
        if self.cfg.num_fcs > 0:
            for fc in self.fcs:
                x = fc(x)

        embeddings = self.fc_embed(x).split([len(p) for p in proposals])
        return embeddings, targets



"""Tracking model utils."""
from typing import List, Tuple

import torch
from torch.nn import functional as F

from vist.struct import InputSample


def split_key_ref_inputs(
    batched_input_samples: List[List[InputSample]],
) -> Tuple[List[InputSample], List[List[InputSample]]]:
    """Split key / ref frame inputs from batched List of InputSample."""
    key_indices = []  # type: List[int]
    ref_indices = []  # type: List[List[int]]
    for input_samples in batched_input_samples:
        curr_ref_indices = list(range(0, len(input_samples)))
        for i, sample in enumerate(input_samples):
            if (
                sample.metadata.attributes is not None
                and sample.metadata.attributes.get("keyframe", False)
            ):
                key_indices.append(curr_ref_indices.pop(i))
                ref_indices.append(curr_ref_indices)
                break

    key_inputs = [
        inputs[key_index]
        for inputs, key_index in zip(batched_input_samples, key_indices)
    ]
    ref_inputs = [
        [inputs[i] for i in curr_ref_indices]
        for inputs, curr_ref_indices in zip(batched_input_samples, ref_indices)
    ]
    return key_inputs, ref_inputs


def cosine_similarity(
    key_embeds: torch.Tensor,
    ref_embeds: torch.Tensor,
    normalize: bool = True,
    temperature: float = -1,
) -> torch.Tensor:
    """Calculate cosine similarity."""
    if normalize:
        key_embeds = F.normalize(key_embeds, p=2, dim=1)
        ref_embeds = F.normalize(ref_embeds, p=2, dim=1)

    dists = torch.mm(key_embeds, ref_embeds.t())

    if temperature > 0:
        dists /= temperature  # pragma: no cover
    return dists



"""Quasi-dense embedding similarity based graph."""
from typing import Dict, List, Optional, Tuple, Union

import torch
from pydantic import validator

from vist.common.bbox.utils import compute_iou
from vist.struct import Boxes2D

from .base import BaseTrackGraph, TrackGraphConfig


class QDTrackGraphConfig(TrackGraphConfig):
    """Quasi-dense similarity based graph config."""

    keep_in_memory: int  # threshold for keeping occluded objects in memory
    init_score_thr: float = 0.7
    obj_score_thr: float = 0.3
    match_score_thr: float = 0.5
    memo_backdrop_frames: int = 1
    memo_momentum: float = 0.8
    nms_conf_thr: float = 0.5
    nms_backdrop_iou_thr: float = 0.3
    nms_class_iou_thr: float = 0.7
    with_cats: bool = True

    @validator("memo_momentum", check_fields=False)
    def validate_memo_momentum(  # pylint: disable=no-self-argument,no-self-use,line-too-long
        cls, value: float
    ) -> float:
        """Check memo_momentum attribute."""
        if not 0 <= value <= 1.0:
            raise ValueError("memo_momentum must be >= 0 and <= 1.0")
        return value

    @validator("keep_in_memory", check_fields=False)
    def validate_keep_in_memory(  # pylint: disable=no-self-argument,no-self-use,line-too-long
        cls, value: int
    ) -> int:
        """Check keep_in_memory attribute."""
        if not value >= 0:
            raise ValueError("keep_in_memory must be >= 0")
        return value

    @validator("memo_backdrop_frames", check_fields=False)
    def validate_memo_backdrop_frames(  # pylint: disable=no-self-argument,no-self-use,line-too-long
        cls, value: int
    ) -> int:
        """Check memo_backdrop_frames attribute."""
        if not value >= 0:
            raise ValueError("memo_backdrop_frames must be >= 0")
        return value


class QDTrackGraph(BaseTrackGraph):
    """Tracking graph construction for quasi-dense instance similarity."""

    def __init__(self, cfg: TrackGraphConfig) -> None:
        """Init."""
        super().__init__(cfg)
        self.cfg = QDTrackGraphConfig(**cfg.dict())

    def reset(self) -> None:
        """Reset tracks."""
        super().reset()
        self.backdrops: List[Dict[str, Union[Boxes2D, torch.Tensor]]] = []

    def get_tracks(
        self,
        device: torch.device,
        frame_id: Optional[int] = None,
        add_backdrops: bool = False,
    ) -> Tuple[Boxes2D, torch.Tensor]:
        """Get active tracks at given frame.

        If frame_id is None, return all tracks in memory.
        """
        bboxs, embeds, cls, ids = [], [], [], []
        for k, v in self.tracks.items():
            if frame_id is None or v["last_frame"] == frame_id:
                bboxs.append(v["bbox"].unsqueeze(0))
                embeds.append(v["embed"].unsqueeze(0))
                cls.append(v["class_id"])
                ids.append(k)

        bboxs = (
            torch.cat(bboxs)
            if len(bboxs) > 0
            else torch.empty((0, 5), device=device)
        )
        embeds = (
            torch.cat(embeds)
            if len(embeds) > 0
            else torch.empty((0,), device=device)
        )
        cls = (
            torch.cat(cls)
            if len(cls) > 0
            else torch.empty((0,), device=device)
        )
        ids = torch.tensor(ids).to(device)

        if add_backdrops:
            for backdrop in self.backdrops:
                backdrop_ids = torch.full(
                    (len(backdrop["embeddings"]),),
                    -1,
                    dtype=torch.long,
                    device=device,
                )
                ids = torch.cat([ids, backdrop_ids])
                bboxs = torch.cat([bboxs, backdrop["detections"].boxes])
                embeds = torch.cat([embeds, backdrop["embeddings"]])
                cls = torch.cat([cls, backdrop["detections"].class_ids])

        return Boxes2D(bboxs, cls, ids), embeds

    def remove_duplicates(
        self, detections: Boxes2D, embeddings: torch.Tensor
    ) -> Tuple[Boxes2D, torch.Tensor]:
        """Remove overlapping objects across classes via nms."""
        # duplicate removal for potential backdrops and cross classes
        _, inds = detections.boxes[:, -1].sort(descending=True)
        detections, embeddings = detections[inds], embeddings[inds]
        valids = embeddings.new_ones((len(detections),))
        ious = compute_iou(detections, detections)
        for i in range(1, len(detections)):
            if detections.boxes[i, -1] < self.cfg.obj_score_thr:
                thr = self.cfg.nms_backdrop_iou_thr
            else:
                thr = self.cfg.nms_class_iou_thr

            if (ious[i, :i] > thr).any():
                valids[i] = 0
        valids = valids == 1
        detections = detections[valids, :]
        embeddings = embeddings[valids, :]
        return detections, embeddings

    def forward(  # type: ignore # pylint: disable=arguments-differ
        self, detections: Boxes2D, frame_id: int, embeddings: torch.Tensor
    ) -> Boxes2D:
        """Process inputs, match detections with existing tracks."""
        detections, embeddings = self.remove_duplicates(detections, embeddings)

        # init ids container
        ids = torch.full(
            (len(detections),), -1, dtype=torch.long, device=detections.device
        )

        # match if buffer is not empty
        if len(detections) > 0 and not self.empty:
            memo_dets, memo_embeds = self.get_tracks(
                detections.device, add_backdrops=True
            )

            # match using bisoftmax metric
            feats = torch.mm(embeddings, memo_embeds.t())
            d2t_scores = feats.softmax(dim=1)
            t2d_scores = feats.softmax(dim=0)
            scores = (d2t_scores + t2d_scores) / 2

            if self.cfg.with_cats:
                cat_same = detections.class_ids.view(
                    -1, 1
                ) == memo_dets.class_ids.view(1, -1)
                scores *= cat_same.float()

            for i in range(len(detections)):
                conf, memo_ind = torch.max(scores[i, :], dim=0)
                cur_id = memo_dets.track_ids[memo_ind]
                if conf > self.cfg.match_score_thr:
                    if cur_id > -1:
                        if detections.boxes[i, -1] > self.cfg.obj_score_thr:
                            ids[i] = cur_id
                            scores[:i, memo_ind] = 0
                            scores[(i + 1) :, memo_ind] = 0
                        elif conf > self.cfg.nms_conf_thr:  # pragma: no cover
                            ids[i] = -2
        new_inds = (ids == -1) & (
            detections.boxes[:, -1] > self.cfg.init_score_thr
        )
        num_news = new_inds.sum()
        ids[new_inds] = torch.arange(
            self.num_tracks,
            self.num_tracks + num_news,
            dtype=torch.long,
            device=ids.device,
        )
        self.num_tracks += num_news

        self.update(ids, detections, embeddings, frame_id)
        result, _ = self.get_tracks(detections.device, frame_id)
        return result

    def update(  # type: ignore # pylint: disable=arguments-differ
        self,
        ids: torch.Tensor,
        detections: Boxes2D,
        embeddings: torch.Tensor,
        frame_id: int,
    ) -> None:
        """Update track memory using matched detections."""
        tracklet_inds = ids > -1

        # update memo
        for cur_id, det, embed in zip(
            ids[tracklet_inds],
            detections[tracklet_inds],
            embeddings[tracklet_inds],
        ):  # type: ignore
            cur_id = int(cur_id)
            if cur_id in self.tracks.keys():
                self.update_track(cur_id, det, embed, frame_id)
            else:
                self.create_track(cur_id, det, embed, frame_id)

        backdrop_inds = torch.nonzero(ids == -1, as_tuple=False).squeeze(1)
        ious = compute_iou(detections[backdrop_inds], detections)
        for i, ind in enumerate(backdrop_inds):
            if (ious[i, :ind] > self.cfg.nms_backdrop_iou_thr).any():
                backdrop_inds[i] = -1
        backdrop_inds = backdrop_inds[backdrop_inds > -1]

        self.backdrops.insert(
            0,
            dict(
                detections=detections[backdrop_inds],
                embeddings=embeddings[backdrop_inds],
            ),
        )

        # delete invalid tracks from memory
        invalid_ids = []
        for k, v in self.tracks.items():
            if frame_id - v["last_frame"] >= self.cfg.keep_in_memory:
                invalid_ids.append(k)
        for invalid_id in invalid_ids:
            self.tracks.pop(invalid_id)

        if len(self.backdrops) > self.cfg.memo_backdrop_frames:
            self.backdrops.pop()

    def update_track(
        self,
        track_id: int,
        detection: Boxes2D,
        embedding: torch.Tensor,
        frame_id: int,
    ) -> None:
        """Update a specific track with a new models."""
        bbox, cls = detection.boxes[0], detection.class_ids[0]
        velocity = (bbox - self.tracks[track_id]["bbox"]) / (
            frame_id - self.tracks[track_id]["last_frame"]
        )
        self.tracks[track_id]["bbox"] = bbox
        self.tracks[track_id]["embed"] = (
            1 - self.cfg.memo_momentum
        ) * self.tracks[track_id]["embed"] + self.cfg.memo_momentum * embedding
        self.tracks[track_id]["last_frame"] = frame_id
        self.tracks[track_id]["class_id"] = cls
        self.tracks[track_id]["velocity"] = (
            self.tracks[track_id]["velocity"]
            * self.tracks[track_id]["acc_frame"]
            + velocity
        ) / (self.tracks[track_id]["acc_frame"] + 1)
        self.tracks[track_id]["acc_frame"] += 1

    def create_track(
        self,
        track_id: int,
        detection: Boxes2D,
        embedding: torch.Tensor,
        frame_id: int,
    ) -> None:
        """Create a new track from a models."""
        bbox, cls = detection.boxes[0], detection.class_ids[0]
        self.tracks[track_id] = dict(
            bbox=bbox,
            embed=embedding,
            class_id=cls,
            last_frame=frame_id,
            velocity=torch.zeros_like(bbox),
            acc_frame=0,
        )



"""VisT module registry."""
from typing import Any, Dict, Optional, Tuple


class RegistryHolder(type):
    """Registry for all modules in VisT."""

    REGISTRY: Dict[str, "RegistryHolder"] = {}

    # Ignore mcs vs. cls since it conflicts with PEP8:
    # https://github.com/PyCQA/pylint/issues/2028
    def __new__(  # type: ignore # pylint: disable=bad-mcs-classmethod-argument
        cls, name: str, bases: Tuple[Any], attrs: Dict[str, Any]
    ) -> "RegistryHolder":
        """Method called when constructing a new class.

        Adds the new class to the detect registry,
        with its module + name as key.
        """
        new_cls = type.__new__(cls, name, bases, attrs)
        assert isinstance(new_cls, RegistryHolder)
        if len(bases):  # must inherit from some base class beyond Registry
            base = str(bases[0]).replace("<class '", "").replace("'>", "")
            module_name = ".".join([*base.split(".")[:-2], new_cls.__name__])
            cls.REGISTRY[module_name] = new_cls
        return new_cls

    @classmethod
    def get_registry(  # pylint: disable=bad-mcs-classmethod-argument
        cls, cls_type: Optional["RegistryHolder"] = None
    ) -> Dict[str, "RegistryHolder"]:
        """Get registered classes, optionally for a specific scope.

        Args:
            cls_type: The super class for which you'd like to get the
            registered subclasses. E.g. input vist.model.BaseModel to get
            all registered models.

        Returns:
            Dict[str, RegistryHolder]: A dictionary with class names as keys
            and classes as values.
        """
        if cls_type is not None:
            return {
                k.split(".")[-1]: v
                for k, v in cls.REGISTRY.items()
                if issubclass(v, cls_type)
            }

        return dict(cls.REGISTRY)  # pragma: no cover



"""Package setup."""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="vist",
    version="0.0",
    author="CVL @ ETHZ",
    author_email="i@yf.io",
    description="VisT Python Package for perception and motion understanding",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://cv.ethz.ch/",
    project_urls={
        "Documentation": "https://github.com/syscv/",
        "Source": "https://github.com/syscv/",
        "Tracker": "https://github.com/syscv/",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "numpy",
        "Pillow",
        "plyfile",
        "psutil",
        "pycocotools",
        "pydantic",
        "pytoml",
        "PyYAML",
        "requests",
        "scalabel",
        "torch",
        "torchvision",
        "tqdm",
        "devtools",
    ],
    package_data={
        "vist": [
            "data/datasets/motchallenge.toml",
            "data/datasets/waymo.toml",
            "data/datasets/kitti.toml",
            "py.typed",
        ]
    },
    include_package_data=True,
)



"""Test cases for parsing config."""
import unittest
from argparse import Namespace

from ..unittest.utils import get_test_file
from .config import parse_config, read_config


class TestLoadConfig(unittest.TestCase):
    """Test cases for vist config parsing."""

    def test_det_yaml(self) -> None:
        """Check models configuration in yaml format."""
        config = read_config(get_test_file("config-det.yaml"))
        self.assertEqual(config.model.type, "test-model")
        self.assertEqual(config.solver.base_lr, 0.02)
        self.assertEqual(config.solver.lr_policy, "step")

    def test_det_toml(self) -> None:
        """Check models configuration in toml format."""
        config = read_config(get_test_file("config-det.toml"))
        self.assertEqual(config.model.type, "test-model")
        self.assertEqual(config.solver.base_lr, 0.02)
        self.assertEqual(config.solver.lr_policy, "step")

    def test_det_args(self) -> None:
        """Check cmd line argument parsing to launch cfg."""
        args = Namespace(
            config=get_test_file("config-det.yaml"),
            device="cuda",
            num_gpus=2,
            cfg_options="dataloader.workers_per_gpu=2",
        )
        cfg = parse_config(args)
        self.assertEqual(cfg.launch.num_gpus, 2)
        self.assertEqual(cfg.dataloader.workers_per_gpu, 2)

    def test_det_notsupported(self) -> None:
        """Check models configuration in not-supported format."""
        self.assertRaises(NotImplementedError, read_config, "")



"""Test cases for max iou matcher."""
import unittest

import torch

from vist.struct import Boxes2D
from vist.unittest.utils import generate_dets

from .base import MatcherConfig
from .max_iou import MaxIoUMatcher


class TestRandom(unittest.TestCase):
    """Test cases for max iou matcher."""

    def test_match(self) -> None:
        """Testcase for sample function."""
        num_boxes = 10

        boxes = [generate_dets(128, 128, num_boxes)]
        matcher = MaxIoUMatcher(
            MatcherConfig(
                type="MaxIoUMatcher",
                thresholds=[0.3, 0.5],
                labels=[0, -1, 1],
                allow_low_quality_matches=False,
            )
        )
        match_result = matcher.match(boxes, boxes)[0]
        self.assertTrue(
            match_result.assigned_gt_indices.numpy().tolist()
            == list(range(num_boxes))
        )

        match_result = matcher.match(boxes, [Boxes2D(torch.empty(0, 5))])[0]
        self.assertTrue((match_result.assigned_labels == 0.0).all())



"""Load data from custom directory."""
import os

from PIL import Image
from scalabel.label.typing import Config, Dataset, Frame, ImageSize

from .base import BaseDatasetLoader


class Custom(BaseDatasetLoader):
    """Custom dataloading class."""

    def load_dataset(self) -> Dataset:
        """Convert data in directory to scalabel format."""
        assert self.cfg.annotations is None and self.cfg.config_path is None

        frames = []
        sub_dirs = list(os.walk(self.cfg.data_root))
        source_dir = os.path.join(self.cfg.data_root, "")  # add trailing slash
        for (root, dirs, files) in sub_dirs:
            if not dirs:
                video_name = os.path.join(root, "").replace(source_dir, "")
                img_files = sorted(
                    [
                        f
                        for f in files
                        if ".jpg" in f or ".png" in f or ".jpeg" in f
                    ]
                )
                if len(img_files) == 0:
                    continue  # pragma: no cover
                for i, img_file in enumerate(img_files):
                    img = Image.open(
                        os.path.join(source_dir, video_name, img_file)
                    )
                    size = ImageSize(width=img.size[0], height=img.size[1])
                    frame = Frame(
                        name=img_file,
                        video_name=video_name,
                        frame_index=i,
                        size=size,
                    )
                    frames.append(frame)

        assert len(frames) > 0, "Input folder didn't contain any images!"
        metadata_cfg = Config(categories=[])
        return Dataset(frames=frames, config=metadata_cfg)



"""Config definitions."""
import os
import sys
from argparse import Namespace
from datetime import datetime
from typing import Any, List, Optional

import toml
import yaml
from pydantic import BaseModel, validator

from vist.data.dataset_mapper import DataloaderConfig
from vist.data.datasets.base import BaseDatasetConfig
from vist.model import BaseModelConfig
from vist.struct import DictStrAny


class Solver(BaseModel):
    """Config for solver."""

    images_per_gpu: int
    lr_policy: str
    base_lr: float
    steps: Optional[List[int]]
    max_iters: int
    warmup_iters: Optional[int]
    checkpoint_period: Optional[int]
    log_period: Optional[int]
    eval_period: Optional[int]


class Launch(BaseModel):
    """Launch configuration.

    Standard Options (command line only):
    action (positional argument): train / predict routine
    config: Filepath to config file

    Launch Options:
    device: Device to train on (cpu / cuda / ..)
    weights: Filepath for weights to load. Set to "detectron2" If you want to
            load weights from detectron2 for a corresponding detector.
    num_gpus: number of gpus per machine
    num_machines: total number of machines
    machine_rank: the rank of this machine (unique per machine)
    dist_url: initialization URL for pytorch distributed backend. See
        https://pytorch.org/docs/stable/distributed.html for details.
    resume: Whether to attempt to resume from the checkpoint directory.
    input_dir: Input directory in case you want to run inference on a folder
    with input data (e.g. images that can be temporally sorted by name)
    output_dir: Specific directory to save checkpoints, logs, etc.
    Default: vist-workspace/<model_name>/<timestamp>
    visualize: If you're running in predict mode, this option lets you
    visualize the model predictions in the output_dir.
    seed: Set random seed for numpy, torch, python. Default: -1,
    i.e. no specific random seed is chosen.
    """

    action: str = ""
    device: str = "cpu"
    weights: Optional[str] = None
    num_gpus: int = 1
    num_machines: int = 1
    machine_rank: int = 0
    # PyTorch still may leave orphan processes in multi-gpu training.
    # Therefore we use a deterministic way to obtain port,
    # so that users are aware of orphan processes by seeing the port occupied.
    port = (
        2 ** 15
        + 2 ** 14
        + hash(os.getuid() if sys.platform != "win32" else 1) % 2 ** 14
    )
    dist_url: str = "tcp://127.0.0.1:{}".format(port)
    resume: bool = False
    input_dir: Optional[str]
    output_dir: str = ""
    visualize: bool = False
    seed: int = -1

    @validator("input_dir", always=True)
    def validate_input_dir(  # pylint: disable=no-self-argument,no-self-use
        cls, value: Optional[str]
    ) -> Optional[str]:
        """Check if input dir exists."""
        if value is not None:
            if not os.path.exists(value):
                raise FileNotFoundError(
                    f"Input directory does not exist: {value}"
                )
        return value


class Config(BaseModel):
    """Overall config object."""

    model: BaseModelConfig
    solver: Solver
    dataloader: DataloaderConfig
    train: List[BaseDatasetConfig] = []
    test: List[BaseDatasetConfig] = []
    launch: Launch = Launch()

    def __init__(self, **data: Any) -> None:  # type: ignore
        """Init config."""
        super().__init__(**data)
        if self.launch.output_dir == "":
            timestamp = (
                str(datetime.now())
                .split(".", maxsplit=1)[0]
                .replace(" ", "_")
                .replace(":", "-")
            )
            self.launch.output_dir = os.path.join(
                "vist-workspace", self.model.type, timestamp
            )
        os.makedirs(self.launch.output_dir, exist_ok=True)


def parse_config(args: Namespace) -> Config:
    """Read config, parse cmd line arguments, create workspace dir."""
    cfg = read_config(args.config)
    for attr, value in args.__dict__.items():
        if attr in Launch.__fields__ and value is not None:
            setattr(cfg.launch, attr, getattr(args, attr))

    if cfg.launch.device == "cpu":
        cfg.launch.num_gpus = 0

    if args.__dict__.get("cfg_options", "") != "":
        cfg_dict = cfg.dict()
        options = args.cfg_options.split(",")
        for option in options:
            key, value = option.split("=")
            keylist_update(cfg_dict, key.split("."), value)
        cfg = Config(**cfg_dict)
    return cfg


def load_config(filepath: str) -> DictStrAny:
    """Load config from file to dict."""
    ext = os.path.splitext(filepath)[1]
    if ext == ".yaml":
        with open(filepath, "r") as f:
            config_dict = yaml.load(f.read(), Loader=yaml.Loader)
    elif ext == ".toml":
        config_dict = toml.load(filepath)
    else:
        raise NotImplementedError(f"Config type {ext} not supported")
    return config_dict  # type: ignore


def read_config(filepath: str) -> Config:
    """Read config file and parse it into Config object.

    The config file can be in yaml or toml.
    toml is recommended for readability.
    """
    config_dict = load_config(filepath)
    if "config" in config_dict:
        cwd = os.getcwd()
        os.chdir(os.path.dirname(filepath))
        subconfig_dict = dict()  # type: DictStrAny
        for cfg in config_dict["config"]:
            assert "path" in cfg, "Config arguments must have path!"
            nested_update(subconfig_dict, load_config(cfg["path"]))

        nested_update(subconfig_dict, config_dict)
        config_dict = subconfig_dict
        os.chdir(cwd)

    config_dict = check_for_dicts(config_dict)
    return Config(**config_dict)


def keylist_update(  # type: ignore
    my_dict: DictStrAny, key_list: List[str], value: Any
) -> None:
    """Update nested dict based on multiple keys saved in a list."""
    cur_key = key_list.pop(0)
    if len(key_list) == 0:
        my_dict[cur_key] = value
        return
    keylist_update(my_dict[cur_key], key_list, value)


def nested_update(ori: DictStrAny, new: DictStrAny) -> DictStrAny:
    """Update function for updating a nested dict."""
    for k, v in new.items():
        if isinstance(v, dict):
            ori[k] = nested_update(ori.get(k, {}), v)
        else:
            ori[k] = v
    return ori


def check_for_dicts(obj: Any) -> Any:  # type: ignore
    """Fix pickle error with a class not being serializable.

    TomlDecoder.get_empty_inline_table.<locals>.DynamicInlineTableDict
    """
    if isinstance(obj, dict):
        return {k: check_for_dicts(v) for k, v in obj.items()}
    return obj



"""OpenMT Label data structures."""
from typing import Dict, List, Optional, Tuple, Type, TypeVar, Union

import torch
from scalabel.label.typing import Box2D, Box3D, Label

from .structures import DataInstance, LabelInstance

TBoxes = TypeVar("TBoxes", bound="Boxes")


class Boxes(DataInstance):
    """Abstract container for 2D / BEV / 3D / ... Boxes.

    boxes: torch.FloatTensor: (N, M) N elements of boxes with M parameters
    class_ids: torch.IntTensor: (N,) where each entry is the class id of
    the respective box.
    track_ids: torch.IntTensor (N,) where each entry is the track id of
    the respective box.
    """

    def __init__(
        self,
        boxes: torch.Tensor,
        class_ids: torch.Tensor = None,
        track_ids: torch.Tensor = None,
        metadata: Optional[Dict[str, Union[bool, int, float, str]]] = None,
    ) -> None:
        """Init."""
        assert isinstance(boxes, torch.Tensor) and len(boxes.shape) == 2
        if class_ids is not None:
            assert isinstance(class_ids, torch.Tensor)
            assert len(boxes) == len(class_ids)
            assert boxes.device == class_ids.device
        if track_ids is not None:
            assert isinstance(track_ids, torch.Tensor)
            assert len(boxes) == len(track_ids)
            assert boxes.device == track_ids.device

        self.boxes = boxes
        self.class_ids = class_ids
        self.track_ids = track_ids
        self.metadata = metadata

    def __getitem__(self: "TBoxes", item) -> "TBoxes":  # type: ignore
        """Shadows tensor based indexing while returning new Boxes."""
        if isinstance(item, tuple):
            item = item[0]
        boxes = self.boxes[item]
        class_ids = (
            self.class_ids[item] if self.class_ids is not None else None
        )
        track_ids = (
            self.track_ids[item] if self.track_ids is not None else None
        )
        if len(boxes.shape) < 2:
            if class_ids is not None:
                class_ids = class_ids.view(1, -1)
            if track_ids is not None:
                track_ids = track_ids.view(1, -1)
            return type(self)(
                boxes.view(1, -1), class_ids, track_ids, self.metadata
            )

        return type(self)(boxes, class_ids, track_ids, self.metadata)

    def __len__(self) -> int:
        """Get length of the object."""
        return len(self.boxes)

    def clone(self: "TBoxes") -> "TBoxes":
        """Create a copy of the object."""
        class_ids = (
            self.class_ids.clone() if self.class_ids is not None else None
        )
        track_ids = (
            self.track_ids.clone() if self.track_ids is not None else None
        )
        return type(self)(
            self.boxes.clone(), class_ids, track_ids, self.metadata
        )

    def to(self: "TBoxes", device: torch.device) -> "TBoxes":
        """Move data to given device."""
        class_ids = (
            self.class_ids.to(device=device)
            if self.class_ids is not None
            else None
        )
        track_ids = (
            self.track_ids.to(device=device)
            if self.track_ids is not None
            else None
        )
        return type(self)(
            self.boxes.to(device=device), class_ids, track_ids, self.metadata
        )

    @classmethod
    def cat(cls: Type["TBoxes"], instances: List["TBoxes"]) -> "TBoxes":
        """Concatenates a list of Boxes into a single Boxes.

        If the Boxes instances have different number of parameters per entry,
        this function will take the minimum and cut additional parameters
        from the other instances.
        """
        assert isinstance(instances, (list, tuple))
        assert len(instances) > 0
        assert all((isinstance(inst, Boxes) for inst in instances))

        boxes, class_ids, track_ids = [], [], []
        has_class_ids = all((b.class_ids is not None for b in instances))
        has_track_ids = all((b.track_ids is not None for b in instances))
        min_param = min((b.boxes.shape[-1] for b in instances))
        for b in instances:
            boxes.append(b.boxes[:, :min_param])
            if has_class_ids:
                class_ids.append(b.class_ids)
            if has_track_ids:
                track_ids.append(b.track_ids)

        cat_boxes = cls(
            torch.cat(boxes),
            torch.cat(class_ids) if has_class_ids else None,
            torch.cat(track_ids) if has_track_ids else None,
        )
        return cat_boxes

    @property
    def device(self) -> torch.device:
        """Get current device of data."""
        return self.boxes.device


class Boxes2D(Boxes, LabelInstance):
    """Container class for 2D boxes.

    boxes: torch.FloatTensor: (N, [4, 5]) where each entry is defined by
    [x1, y1, x2, y2, Optional[score]]
    class_ids: torch.LongTensor: (N,) where each entry is the class id of
    the respective box.
    track_ids: torch.LongTensor (N,) where each entry is the track id of
    the respective box.
    """

    def scale(self, scale_factor_xy: Tuple[float, float]) -> None:
        """Scale bounding boxes according to factor."""
        self.boxes[:, [0, 2]] *= scale_factor_xy[0]
        self.boxes[:, [1, 3]] *= scale_factor_xy[1]

    def clip(self, image_wh: Tuple[float, float]) -> None:
        """Clip bounding boxes according to image_wh."""
        self.boxes[:, [0, 2]] = self.boxes[:, [0, 2]].clamp(0, image_wh[0] - 1)
        self.boxes[:, [1, 3]] = self.boxes[:, [1, 3]].clamp(0, image_wh[1] - 1)

    @classmethod
    def from_scalabel(
        cls,
        labels: List[Label],
        class_to_idx: Dict[str, int],
        label_id_to_idx: Optional[Dict[str, int]] = None,
    ) -> "Boxes2D":
        """Convert from scalabel format to internal."""
        box_list, cls_list, idx_list = [], [], []
        has_class_ids = all((b.category is not None for b in labels))
        for i, label in enumerate(labels):
            box, score, box_cls, l_id = (
                label.box2d,
                label.score,
                label.category,
                label.id,
            )
            if box is None:
                continue

            if score is None:
                box_list.append([box.x1, box.y1, box.x2, box.y2])
            else:
                box_list.append([box.x1, box.y1, box.x2, box.y2, score])

            if has_class_ids:
                cls_list.append(class_to_idx[box_cls])  # type: ignore
            idx = label_id_to_idx[l_id] if label_id_to_idx is not None else i
            idx_list.append(idx)

        box_tensor = torch.tensor(box_list, dtype=torch.float32)
        class_ids = (
            torch.tensor(cls_list, dtype=torch.long) if has_class_ids else None
        )
        track_ids = torch.tensor(idx_list, dtype=torch.long)
        return Boxes2D(box_tensor, class_ids, track_ids)

    def to_scalabel(self, idx_to_class: Dict[int, str]) -> List[Label]:
        """Convert from internal to scalabel format."""
        labels = []
        for i in range(len(self.boxes)):
            if self.track_ids is not None:
                label_id = str(self.track_ids[i].item())
            else:
                label_id = str(i)
            box = Box2D(
                x1=float(self.boxes[i, 0]),
                y1=float(self.boxes[i, 1]),
                x2=float(self.boxes[i, 2]),
                y2=float(self.boxes[i, 3]),
            )
            if self.boxes.shape[-1] == 5:
                score = float(self.boxes[i, 4])  # type: Optional[float]
            else:
                score = None
            label_dict = dict(id=label_id, box2d=box, score=score)

            cls = idx_to_class[int(self.class_ids[i])]
            label_dict["category"] = cls
            labels.append(Label(**label_dict))

        return labels


class Boxes3D(Boxes, LabelInstance):
    """Container class for 3D boxes.

    boxes: torch.FloatTensor: (N, [7, 8]) where each entry is defined as
    [x, y, z, h, w, l, ry, Optional[score]] or (N, [9, 10]) where each entry
    is defined by [x, y, z, h, w, l, rx, ry, rz, Optional[score]].
    class_ids: torch.LongTensor: (N,) where each entry is the class id of
    the respective box.
    track_ids: torch.LongTensor (N,) where each entry is the track id of
    the respective box.

    x,y,z are in OpenCV camera coordinate system. l, h, w, are the 3D box
    dimensions and correspond to their respective axis (length first (x),
    height second (y), width last (z). The rotations are axis angles w.r.t.
    each axis (x,y,z).
    """

    @classmethod
    def from_scalabel(
        cls,
        labels: List[Label],
        class_to_idx: Dict[str, int],
        label_id_to_idx: Optional[Dict[str, int]] = None,
    ) -> "Boxes3D":
        """Convert from scalabel format to internal."""
        box_list, cls_list, idx_list = [], [], []
        has_class_ids = all((b.category is not None for b in labels))
        for i, label in enumerate(labels):
            box, score, box_cls, l_id = (
                label.box3d,
                label.score,
                label.category,
                label.id,
            )
            if box is None:
                continue

            if score is None:
                box_list.append(
                    [*box.location, *box.dimension, *box.orientation]
                )
            else:
                box_list.append(
                    [*box.location, *box.dimension, *box.orientation, score]
                )
            if has_class_ids:
                cls_list.append(class_to_idx[box_cls])  # type: ignore
            idx = label_id_to_idx[l_id] if label_id_to_idx is not None else i
            idx_list.append(idx)

        box_tensor = torch.tensor(box_list, dtype=torch.float32)
        class_ids = (
            torch.tensor(cls_list, dtype=torch.long) if has_class_ids else None
        )
        track_ids = torch.tensor(idx_list, dtype=torch.long)
        return Boxes3D(box_tensor, class_ids, track_ids)

    def to_scalabel(self, idx_to_class: Dict[int, str]) -> List[Label]:
        """Convert from internal to scalabel format."""
        labels = []
        for i in range(len(self.boxes)):
            if self.track_ids is not None:
                label_id = str(self.track_ids[i].item())
            else:
                label_id = str(i)

            if self.boxes.shape[-1] < 9:
                rx = 0.0
                ry = float(self.boxes[i, 6])
                rz = 0.0
                if self.boxes.shape[-1] == 8:
                    score = float(self.boxes[i, 7])  # type: Optional[float]
                else:
                    score = None
            else:
                rx = float(self.boxes[i, 6])
                ry = float(self.boxes[i, 7])
                rz = float(self.boxes[i, 8])
                if self.boxes.shape[-1] == 10:
                    score = float(self.boxes[i, 9])
                else:
                    score = None

            box = Box3D(
                location=[
                    float(self.boxes[i, 0]),
                    float(self.boxes[i, 1]),
                    float(self.boxes[i, 2]),
                ],
                dimension=[
                    float(self.boxes[i, 3]),
                    float(self.boxes[i, 4]),
                    float(self.boxes[i, 5]),
                ],
                orientation=[rx, ry, rz],
                alpha=-1.0,
            )

            label_dict = dict(id=label_id, box3d=box, score=score)

            cls = idx_to_class[int(self.class_ids[i])]
            label_dict["category"] = cls
            labels.append(Label(**label_dict))

        return labels
